/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.aggregates.cube;

import java.util.*;
import java.util.concurrent.atomic.AtomicLong;

public class Dimensions {
    private static final boolean FLYWEIGHT = false;

    private static final FlyweightSet<Dimensions> cache = FLYWEIGHT ? new FlyweightSet<Dimensions>() : null;

    private static Dimensions get(Object... path) {
        if(FLYWEIGHT) {
            return cache.put(new Dimensions(path));
        } else {
            return new Dimensions(path);
        }
    }

    public final Object[] path;
    public final Map<Object,Integer> columnIndex;
    private final List<OrderedColumn> orderedColumns;

    public Dimensions(Object... path) {
        this.path = path;
        columnIndex = new HashMap<Object, Integer>();
        for (int i = 0; i < path.length; i++) {
            Object column = path[i];
            columnIndex.put(column, i);
        }
        this.orderedColumns = new ArrayList<OrderedColumn>();
        for (int i = 0; i < path.length; i++) {
            Object column = path[i];
            orderedColumns.add(new OrderedColumn(column, i));
        }
        Collections.sort(orderedColumns);
    }

    public static Dimensions create(Object... path) {
        return get(path);
    }

    public Group createOrderedGroup(Group group) {
        Object[] path = new Object[orderedColumns.size()];
        for (int i = 0; i < orderedColumns.size(); i++) {
            OrderedColumn orderedColumn = orderedColumns.get(i);
            path[i] = group.getPath(orderedColumn.getIndex());
        }
        return GroupFactory.create(path);
    }

    public Dimensions drillDown(Object... columns) {
        Object[] path = new Object[this.path.length + columns.length];
        System.arraycopy(this.path, 0, path, 0, this.path.length);
        System.arraycopy(columns, 0, path, this.path.length, columns.length);
        return Dimensions.create(path);
    }

    public Dimensions drillUp() {
        assert this.path.length > 0: this;

        if(this.path.length > 0) {
            Object[] path = new Object[this.path.length - 1];
            System.arraycopy(this.path, 0, path, 0, path.length);
            return Dimensions.create(path);
        } else {
            return null;
        }
    }

    public Object getLast() {
        return path[path.length - 1];
    }

    @Override
    public boolean equals(Object o) {
        if(FLYWEIGHT) {
            return this == o;
        } else {
            if (this == o) return true;
            if (!(o instanceof Dimensions)) return false;

            Dimensions path1 = (Dimensions) o;

            if(InMemoryCube.UNIQUE_CUBOID) {
                return orderedColumns.equals(path1.orderedColumns);
            } else {
                // Probably incorrect - comparing Object[] arrays with Arrays.equals
                return Arrays.equals(path, path1.path);
            }
        }
    }

    @Override
    public int hashCode() {
        if(FLYWEIGHT) {
            return super.hashCode();
        } else {
            if(InMemoryCube.UNIQUE_CUBOID) {
                return orderedColumns.hashCode();
            } else {
                return Arrays.hashCode(path);
            }
        }
    }

    @Override
    public String toString() {
        return "Dimensions{" +
                "path=" + Arrays.toString(path) +
                '}';
    }

    public Integer getIndex(Object column) {
        return columnIndex.get(column);
    }

    private static class OrderedColumn implements Comparable<OrderedColumn> {
        private static final Map<Object,Long> ids = new WeakHashMap<Object, Long>();
        private static final AtomicLong nextId = new AtomicLong(0);

        private static long getId(Object column) {
            if(!ids.containsKey(column)) {
                ids.put(column, nextId.getAndIncrement());
            }
            return ids.get(column);
        }

        private final Object column;
        private final int index;

        private OrderedColumn(Object column, int index) {
            this.column = column;
            this.index = index;
        }

        public Object getColumn() {
            return column;
        }

        public int getIndex() {
            return index;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            OrderedColumn that = (OrderedColumn) o;

            return column.equals(that.column);

        }

        @Override
        public int hashCode() {
            return column.hashCode();
        }

        @Override
        public int compareTo(OrderedColumn o) {
            final long x = getId(getColumn());
            final long y = getId(o.getColumn());
            return  (x < y) ? -1 : ((x == y) ? 0 : 1);
        }
    }
}