/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.aggregates.cube;

import com.macrofocus.application.ApplicationModel;
import com.macrofocus.application.ApplicationModelFactory;
import com.macrofocus.molap.aggregates.aggregation.*;
import com.macrofocus.molap.aggregates.cuboid.ApexCuboid;
import com.macrofocus.molap.aggregates.cuboid.Cuboid;
import com.macrofocus.molap.aggregates.cuboid.DrilledCuboid;
import com.macrofocus.molap.aggregates.query.CuboidQuery;
import com.macrofocus.molap.aggregates.query.Query;
import com.macrofocus.molap.aggregates.query.Structure;
import com.macrofocus.molap.dataframe.DataFrame;

import java.util.HashMap;
import java.util.Map;

/**
 * An in-memory multidimensional cube.
 */
public class InMemoryCube implements Cube {
    public static final boolean UNIQUE_CUBOID = false;

    private final DataFrame dataFrame;

    private final Cuboid apex;
    private final Dimensions root;
    private final Dice dice;
    private final Structure structure;

    private final Map<DimensionsDiceStructure, Cuboid> cuboids = new HashMap<DimensionsDiceStructure, Cuboid>();
    private final Map<Cuboid, Cuboid> child = new HashMap<Cuboid, Cuboid>();

    public InMemoryCube(DataFrame dataFrame) {
        this.dataFrame = dataFrame;

        root = Dimensions.create();
        dice = Dice.create();
        structure = Structure.create();
        apex = new ApexCuboid(this, root);
        cuboids.put(new DimensionsDiceStructure(root, dice, structure), apex);
    }

    @Override
    public DataFrame getDataFrame() {
        return dataFrame;
    }

    @Override
    public Cuboid getApexCuboid() {
        return apex;
    }

    @Override
    public Dimensions getRoot() {
        return root;
    }

    @Override
    public Dice getDice() {
        return dice;
    }

    @Override
    public Structure getStructure() {
        return structure;
    }

    @Override
    public Cuboid getCuboid(Dimensions dimensions, Dice dice, Structure structure) {
        assert dimensions.path.length == dice.path.length && dice.path.length == structure.getPath().length: dimensions + ", " + dice + ", " + structure;

        final DimensionsDiceStructure key = new DimensionsDiceStructure(dimensions, dice, structure);
        if(!cuboids.containsKey(key)) {
//            System.err.println("Creating cuboid for " + dimensions + ", " + dice + ", " + structure);

            final Dimensions drillUp = dimensions.drillUp();
            final Dice diceUp = dice.drillUp();
            final Structure structureUp = structure.drillUp();

            // Recursion
            final Cuboid parentCuboid = getCuboid(drillUp, diceUp, structureUp);
            final DrilledCuboid cuboid = new DrilledCuboid(this, parentCuboid, dimensions, dice, structure);
            cuboids.put(key, cuboid);

            child.put(parentCuboid, cuboid);

            return cuboid;
        }

        return cuboids.get(key);
    }

    @Override
    public Cuboid getChildCuboid(Cuboid cuboid) {
        return child.get(cuboid);
//
//        for (Dimensions dimensions : cuboids.keySet()) {
//            Dimensions up = dimensions.drillUp();
//            if(up != null && up.equals(cuboid.getDimensions())) {
//                return cuboids.get(dimensions);
//            }
//        }
//        return null;
    }

    @Override
    public Query query(Aggregation... aggregations) {
        return new CuboidQuery(getApexCuboid(), aggregations);
    }

    @Override
    public SumAggregation getSum(String name) {
        return new SumAggregation(getDataFrame().getColumn(name));
    }

    @Override
    public CountAggregation getCount(String name) {
        return new CountAggregation(getDataFrame().getColumn(name));
    }

    @Override
    public MinAggregation getMin(String name) {
        return new MinAggregation(getDataFrame().getColumn(name));
    }

    @Override
    public MaxAggregation getMax(String name) {
        return new MaxAggregation(getDataFrame().getColumn(name));
    }

    @Override
    public MeanAggregation getMean(String name) {
        return new MeanAggregation(getDataFrame().getColumn(name));
    }

    @Override
    public VarianceAggregation getVariance(String name) {
        return new VarianceAggregation(getDataFrame().getColumn(name));
    }

    @Override
    public VarianceAggregation getVariance(String name, String population) {
        return new VarianceAggregation(getDataFrame().getColumn(name), getDataFrame().getColumn(population));
    }

    @Override
    public StdDevAggregation getStdDev(String name) {
        return new StdDevAggregation(getDataFrame().getColumn(name));
    }

    @Override
    public CountDistinctAggregation getCountDistinct(String name) {
        return new CountDistinctAggregation(getDataFrame().getColumn(name));
    }

    @Override
    public String toString() {
        return "InMemoryCube{" +
                "cuboids=" + cuboids.keySet() +
                '}';
    }

    private static class DimensionsDiceStructure {
        private final Dimensions dimensions;
        private final Dice dice;
        private final Structure structure;

        private DimensionsDiceStructure(Dimensions dimensions, Dice dice, Structure structure) {
            this.dimensions = dimensions;
            this.dice = dice;
            this.structure = structure;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            DimensionsDiceStructure that = (DimensionsDiceStructure) o;

            if (!dice.equals(that.dice)) return false;
            if (!dimensions.equals(that.dimensions)) return false;
            return structure.equals(that.structure);

        }

        @Override
        public int hashCode() {
            int result = dimensions.hashCode();
            result = 31 * result + dice.hashCode();
            result = 31 * result + structure.hashCode();
            return result;
        }

        @Override
        public String toString() {
            return "DimensionsDiceStructure{" +
                    "dimensions=" + dimensions +
                    ", dice=" + dice +
                    ", structure=" + structure +
                    '}';
        }
    }

    public static void main(String[] args) {
        printUsage(ApplicationModelFactory.MOLAP_V1);
    }

    /**
     * Print the usage message and exit
     */
    public static void printUsage(ApplicationModel applicationModel) {
        String lSep = System.getProperty("line.separator");
        StringBuffer msg = new StringBuffer();
        msg.append(lSep);
        msg.append("------------------------------------------------------------------------ ").append(lSep);
        msg.append(applicationModel.getName()).append(" v. ").append(applicationModel.getVersion()).append(lSep);
        msg.append("Copyright (c) ").append(applicationModel.getCopyright()).append(" Macrofocus GmbH. All rights reserved.").append(lSep);
        msg.append("------------------------------------------------------------------------ ").append(lSep).append(lSep);
        System.out.println(msg.toString());
        System.exit(0);
    }

}
