/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.aggregates.cuboid;

import com.macrofocus.molap.aggregates.cube.Cube;
import com.macrofocus.molap.aggregates.cube.DistributiveStatistics;
import com.macrofocus.molap.aggregates.cube.Group;
import com.macrofocus.molap.series.Series;

import java.util.*;

public abstract class AbstractCuboid implements Cuboid{
    protected final Cube cube;

    private boolean dirty = true;

    private Map<GroupColumn, DistributiveStatistics> distributiveStatisticsCache;
    private Map<GroupColumn, Set<Object>> distinctCache;

    protected AbstractCuboid(Cube cube) {
        this.cube = cube;
    }

    public boolean isDirty() {
        return dirty;
    }

    public void setDirty(boolean dirty) {
        this.dirty = dirty;
    }

    protected void build() {
        distributiveStatisticsCache = Collections.synchronizedMap(new HashMap<GroupColumn, DistributiveStatistics>());
        distinctCache = Collections.synchronizedMap(new HashMap<GroupColumn, Set<Object>>());
    }

    @Override
    public DistributiveStatistics getDistributiveStatistics(Group group, Series column) {
        assert group != null;
        assert column != null;

        if(isDirty()) {
            build();
        }
        final GroupColumn key = new GroupColumn(group, column);
        if(!distributiveStatisticsCache.containsKey(key)) {
            Cuboid child = cube.getChildCuboid(this);
            DistributiveStatistics distributiveStatistics;

            if(child != null) {
                distributiveStatistics = new DistributiveStatistics();
                for(Group childGroup: child.getGroupsStartingWith(group)) {
                    assert childGroup.startsWith(group);

                    DistributiveStatistics childDistributiveStatistics = child.getDistributiveStatistics(childGroup, column);
                    distributiveStatistics.add(childDistributiveStatistics);
                }
            } else {
                final Iterable<Object> rows = getRows(group);
                if(rows != null) {
                    distributiveStatistics = new DistributiveStatistics(column, rows);
                } else {
                    distributiveStatistics = null;
                }
            }
            distributiveStatisticsCache.put(key, distributiveStatistics);
            return distributiveStatistics;
        }
        return distributiveStatisticsCache.get(key);
    }

    @Override
    public Set<Object> getDistinct(Group group, Series column) {
        if(isDirty()) {
            build();
        }
        final GroupColumn key = new GroupColumn(group, column);
        if(!distinctCache.containsKey(key)) {
            Cuboid child = cube.getChildCuboid(this);
            Set<Object> set;

            if(child != null) {
                set = new HashSet<Object>();
                for(Group childGroup: child.getGroupsStartingWith(group)) {
                    assert childGroup.startsWith(group);

                    Set<Object> childSet = child.getDistinct(childGroup, column);
                    set.addAll(childSet);
                }
            } else {
                final Iterable<Object> rows = getRows(group);
                if(rows != null) {
                    set = new HashSet<Object>();
                    for (Object row : rows) {
                        Object value = column.get(row);
                        set.add(value);
                    }

                } else {
                    set = null;
                }
            }
            distinctCache.put(key, set);
            return set;
        }
        return distinctCache.get(key);
    }

    private static class GroupColumn {
        private final Group group;
        private final Series column;

        private GroupColumn(Group group, Series column) {
            assert group != null;
            assert column != null;

            this.group = group;
            this.column = column;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            GroupColumn that = (GroupColumn) o;

            if (!column.equals(that.column)) return false;
            return group.equals(that.group);

        }

        @Override
        public int hashCode() {
            int result = group.hashCode();
            result = 31 * result + column.hashCode();
            return result;
        }

        @Override
        public String toString() {
            return "GroupColumn{" +
                    "group=" + group +
                    ", column=" + column +
                    '}';
        }
    }
}
