/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.aggregates.cuboid;

import com.macrofocus.molap.aggregates.cube.Cube;
import com.macrofocus.molap.aggregates.cube.Dice;
import com.macrofocus.molap.aggregates.cube.Dimensions;
import com.macrofocus.molap.aggregates.cube.Group;
import com.macrofocus.molap.aggregates.query.Structure;
import com.macrofocus.molap.dataframe.DataFrameEvent;
import com.macrofocus.molap.dataframe.DataFrameListener;
import com.macrofocus.molap.series.Series;
import com.macrofocus.utils.ConcurrentHashSet;

import java.util.*;

/**
 * A child cuboid, in principle ultimately rooted on an <code>ApexCuboid</code>.
 *
 * @see com.macrofocus.molap.aggregates.cuboid.ApexCuboid
 */
public class DrilledCuboid extends AbstractCuboid {
    private final Cuboid parent;
    private final Dimensions dimensions;
    private final Dice dice;
    private final Structure structure;

    private Set<Object> values;
    private Map<Group, Set<Group>> parentChildren;
    private Map<Group, List<Object>> rows;

    private final DataFrameListener listener = new DataFrameListener() {
        @Override
        public void dataFrameChanged(DataFrameEvent event) {
            setDirty(true);
        }
    };

    public DrilledCuboid(Cube cube, Cuboid parent, Dimensions dimensions, Dice dice, Structure structure) {
        super(cube);
        this.parent = parent;
        this.dimensions = dimensions;
        this.dice = dice;
        this.structure = structure;

        cube.getDataFrame().addWeakDataFrameListener(listener);
    }

    @Override
    protected void build() {
        super.build();

        Set<Object> values = new LinkedHashSet<Object>();
        Map<Group, Set<Group>> parentChildren = new LinkedHashMap<Group, Set<Group>>();
        Map<Group, List<Object>> rows = new LinkedHashMap<Group, List<Object>>();

        final Object last = dimensions.getLast();
        Series series = cube.getDataFrame().getColumn(last);
        for(final Group parentGroup: parent.getGroups()) {
            final Set<Group> children;
            if(!parentChildren.containsKey(parentGroup)) {
                children = new ConcurrentHashSet<Group>();
                parentChildren.put(parentGroup, children);
            } else {
                children = parentChildren.get(parentGroup);
            }

            for(Object row: parent.getRows(parentGroup)) {
                Object value = series.get(row);
//                Object value = cube.getDataFrame().getValue(row, last);
                if(dice.getLast() == null || dice.getLast().contains(value)) {
                    if(structure.getLast()) {
                        value = Cuboid.ALL;
                    }
                    // Values
                    values.add(value);

                    Group current = parentGroup.drillDown(value);

                    // Children
                    if(!children.contains(current)) {
                        children.add(current);
                    }

                    // Rows
                    final List<Object> r;
                    if(!rows.containsKey(current)) {
                        r = new ArrayList<Object>();
                        rows.put(current, r);
                    } else {
                        r = rows.get(current);
                    }

                    assert r != null;

                    r.add(row);
                }
            }
        }

        this.values = values;
        this.parentChildren = parentChildren;
        this.rows = rows;

        setDirty(false);
    }

    @Override
    public Cuboid drillDown(Object... columns) {
        Set[] diceValues = new Set[columns.length];
        Boolean[] structureValues = new Boolean[columns.length];
        Arrays.fill(structureValues, Boolean.FALSE);
        return cube.getCuboid(dimensions.drillDown(columns), dice.drillDown(diceValues), structure.drillDown(structureValues));
    }

    @Override
    public Cuboid drillUp() {
        return cube.getCuboid(dimensions.drillUp(), dice.drillUp(), structure.drillUp());
    }

    @Override
    public Cuboid slice(Object value) {
        return dice(Collections.singleton(value));
    }

    @Override
    public Cuboid dice(Set<Object> valuesSets) {
        return cube.getCuboid(dimensions, dice.changeLast(valuesSets), structure);
    }

    @Override
    public Cuboid collapse() {
        return cube.getCuboid(dimensions, dice, structure.changeLast(true));
    }

    @Override
    public Cube getCube() {
        return cube;
    }

    @Override
    public Dimensions getDimensions() {
        return dimensions;
    }

    @Override
    public Iterable<Group> getGroups() {
        if(isDirty()) {
            build();
        }
        return rows.keySet();
    }

    @Override
    public Iterable<Group> getGroupsStartingWith(Group group) {
        if(isDirty()) {
            build();
        }
        return parentChildren.get(group);
    }

    @Override
    public int getGroupsStartingWithCount(Group group) {
        if(isDirty()) {
            build();
        }
        return parentChildren.get(group).size();
    }

    @Override
    public int getGroupsCount() {
        if(isDirty()) {
            build();
        }
        return rows.keySet().size();
    }

    @Override
    public Set<Object> getValues() {
        if(isDirty()) {
            build();
        }
        return values;
    }

    @Override
    public List<Object> getRows(Group group) {
        if(isDirty()) {
            build();
        }
        return rows.get(group);
    }

    @Override
    public String toString() {
        return "DrilledCuboid{" +
                "dimensions=" + dimensions +
                '}';
    }
}
