/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.aggregates.query;

import com.macrofocus.molap.adapter.AggregatesTableModel;
import com.macrofocus.molap.aggregates.aggregation.Aggregation;
import com.macrofocus.molap.aggregates.cuboid.Cuboid;
import com.macrofocus.molap.dataframe.DataFrame;

import javax.swing.table.TableModel;
import java.io.PrintStream;
import java.lang.ref.WeakReference;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArrayList;

public abstract class AbstractQuery implements Query {
    private boolean dirty = true;

    private final List<QueryListener> listeners;

    public AbstractQuery() {
        listeners = new CopyOnWriteArrayList<QueryListener>();
    }

    public boolean isDirty() {
        return dirty;
    }

    public void setDirty(boolean dirty) {
        this.dirty = dirty;
    }

    @Override
    public CuboidQuery drillDown(Object... columns) {
        return new CuboidQuery(this, new CuboidQuery.DrillDownOperation(columns), getAggregations());
    }

    @Override
    public Query drillUp() {
        return new CuboidQuery(this, new CuboidQuery.DrillUpOperation(), getAggregations());
    }

    @Override
    public Query slice(Object value) {
        return new CuboidQuery(this, new CuboidQuery.SliceOperation(value), getAggregations());
    }

    @Override
    public Query dice(Set<Object> valuesSets) {
        return new CuboidQuery(this, new CuboidQuery.DiceOperation(valuesSets), getAggregations());
    }

    @Override
    public Query pivot(Aggregation aggregation, Object[] columns) {
        return new PivotQuery(this, aggregation, columns);
    }

    @Override
    public Query pivot(Aggregation aggregation) {
        return new PivotQuery(this, aggregation);
    }

    @Override
    public Query collapse() {
        return new CuboidQuery(this, new CuboidQuery.CollapseOperation(), getAggregations());
    }

    @Override
    public Query order(Aggregation... aggregations) {
        return new OrderQuery(this, aggregations);
    }

    @Override
    public Query on(Cuboid... cuboids) {
        return new CuboidsQuery(cuboids, getAggregations());
    }

    @Override
    public AggregatesTableModel asTableModel() {
        return new AggregatesTableModel(this);
    }

    @Override
    public DataFrame asDataFrame(boolean includeIndex) {
        return new QueryDataFrame(this, includeIndex);
    }

    @Override
    public void printResults(PrintStream out, String caption, boolean html) {
        TableModel tableModel = asTableModel();

        if(!html) {
            if(caption != null) {
                System.out.println(caption);
            }

            int cols = tableModel.getColumnCount();
            for (int i = 0; i < cols; i++) {
                final String o = tableModel.getColumnName(i);
                final String s;
                if(o != null) {
                    s = o;
                } else {
                    s = "n.a.";
                }
                out.printf("%20s", abbreviate(s, 19));
            }
    //        System.out.println();
    //        for (int i = 0; i < cols; i++) {
    //            final String o = tableModel.getColumnClass(i).getSimpleName();
    //            final String s;
    //            if(o != null) {
    //                s = o;
    //            } else {
    //                s = "n.a.";
    //            }
    //            out.printf("%20s", abbreviate(s, 19));
    //        }
            out.println();
            for(int row = 0; row < tableModel.getRowCount(); row++) {
                for (int i = 0; i < cols; i++) {
                    final Object o = tableModel.getValueAt(row, i);
                    final String s;
                    if(o != null) {
                        s = o.toString();
                    } else {
                        s = "n.a.";
                    }
                    out.printf("%20s", abbreviate(s, 19));
                }
                out.println();
            }
        } else {
            out.println("<table style=\"width: auto; border-collapse: separate;\">");

            if(caption != null) {
                out.println("<caption>" + caption + "</caption>");
            }

            out.println("<tbody>");
            out.println("<tr>");
            int cols = tableModel.getColumnCount();
            for (int i = 0; i < cols; i++) {
                final String o = tableModel.getColumnName(i);
                final String s;
                if(o != null) {
                    s = o;
                } else {
                    s = "n.a.";
                }
                out.print("<th>");
                out.print(s);
                out.println("</th>");
            }
            out.println("</tr>");

            for(int row = 0; row < tableModel.getRowCount(); row++) {
                out.println("<tr>");
                for (int i = 0; i < cols; i++) {
                    final Object o = tableModel.getValueAt(row, i);
                    final String s;
                    if(o != null) {
                        s = o.toString();
                    } else {
                        s = "n.a.";
                    }
                    if(o instanceof Number) {
                        out.print("<td align=\"right\">");
                    } else {
                        out.print("<td>");
                    }
                    out.print(s);
                    out.println("</td>");
                }
                out.println("</tr>");
            }
            out.println("</tbody>");

            out.println("</table>");
        }
    }

    private static String abbreviate(String str, int maxWidth) {
        if(str == null) {
            return null;
        }
        if (str.length() <= maxWidth) {
            return str;
        }
        return str.substring(0, maxWidth - 3) + "...";
    }

    @Override
    public void addQueryListener(final QueryListener listener) {
        synchronized (listeners) {
            listeners.add(listener);
        }
    }

    @Override
    public void addWeakQueryListener(final QueryListener listener) {
        synchronized (listeners) {
            final WeakQueryListener weakListener = new WeakQueryListener(listener);
            listeners.add(weakListener);
        }
    }

    @Override
    public void removeQueryListener(final QueryListener listener) {
        synchronized (listeners) {
            if (listener instanceof WeakQueryListener) {
                boolean removed = listeners.remove(listener);
                assert removed : listener;
            } else {
                QueryListener toRemove = null;

                for (QueryListener queryListener : listeners) {
                    QueryListener comparable;
                    if (queryListener instanceof WeakQueryListener) {
                        comparable = ((WeakQueryListener) queryListener).getReference();
                    } else {
                        comparable = queryListener;
                    }

                    if (listener.equals(comparable)) {
                        toRemove = queryListener;
                    }
                }

                if (toRemove != null) {
                    boolean removed = listeners.remove(toRemove);
                    assert removed : listener;
                }
            }
        }
    }

    @Override
    public void removeQueryListeners() {
        synchronized (listeners) {
            listeners.clear();
        }
    }

    protected void notifyQueryChanged() {
        for (final QueryListener listener : listeners) {
            listener.queryChanged();
        }
    }

    private class WeakQueryListener<E> implements QueryListener {
        private final WeakReference<QueryListener> l_ref;

        public WeakQueryListener(final QueryListener listener) {
            this.l_ref = new WeakReference<QueryListener>(listener);
        }

        @Override
        public void queryChanged() {
            final QueryListener l = getReference();
            if (l != null) {
                l.queryChanged();
            } else {
                removeQueryListener((QueryListener) this);
            }
        }

        private QueryListener getReference() {
            return l_ref.get();
        }

        public String toString() {
            final QueryListener l = getReference();
            if (l != null) {
                return "Weak[" + l + "]";
            } else {
                return super.toString();
            }
        }
    }
}
