/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.aggregates.query;

import com.macrofocus.molap.aggregates.aggregation.Aggregation;
import com.macrofocus.molap.aggregates.cube.Dimensions;
import com.macrofocus.molap.aggregates.cube.Group;
import com.macrofocus.molap.aggregates.cuboid.Cuboid;
import com.macrofocus.molap.dataframe.DataFrameEvent;
import com.macrofocus.molap.dataframe.DataFrameListener;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

public class CuboidQuery extends AbstractQuery {
    private Cuboid cuboid;
    private Query query;
    private Operation operation;
    private final Aggregation[] aggregations;

    private List<Group> groups;

    final QueryListener queryListener = new QueryListener() {
        @Override
        public void queryChanged() {
            cuboid = null;
            setDirty(true);
            notifyQueryChanged();
        }
    };

    final DataFrameListener dataFrameListener = new DataFrameListener() {
        @Override
        public void dataFrameChanged(DataFrameEvent event) {
            setDirty(true);
            notifyQueryChanged();
        }
    };

    public CuboidQuery(Cuboid cuboid, Aggregation... aggregations) {
        assert cuboid != null;

        this.cuboid = cuboid;
        this.aggregations = aggregations;

        this.cuboid.getCube().getDataFrame().addWeakDataFrameListener(dataFrameListener);
    }

    public CuboidQuery(Query query, Operation operation, Aggregation... aggregations) {
        assert query != null;

        this.query = query;
        this.operation = operation;
        this.aggregations = aggregations;

        this.query.addWeakQueryListener(queryListener);
    }

    private void run() {
        List<Group> groups = new ArrayList<Group>();
        for (Group group : getCuboid().getGroups()) {
            groups.add(group);
        }
        this.groups = groups;
        setDirty(false);
    }

    @Override
    public Object getValue(Group path, Aggregation aggregation) {
        return aggregation.compute(getCuboid(), path);
    }

    @Override
    public Dimensions getDimensions() {
        return getCuboid().getDimensions();
    }

    @Override
    public Aggregation[] getAggregations() {
        return aggregations;
    }

    @Override
    public List<Group> getGroups() {
        if(isDirty()) {
            run();
        }

        return groups;
    }

    @Override
    public Cuboid getCuboid() {
        if(cuboid == null) {
            cuboid = operation.get(query.getCuboid());
        }
        return cuboid;
    }

    @Override
    public void setDice(Set<Object> valuesSets) {
        operation = new DiceOperation(valuesSets);
        cuboid = null;
        setDirty(true);
        notifyQueryChanged();
    }

    private static interface Operation {
        Cuboid get(Cuboid cuboid);
    }

    public static class DrillDownOperation implements Operation {
        private final Object[] columns;

        public DrillDownOperation(Object... columns) {
            this.columns = columns;
        }

        @Override
        public Cuboid get(Cuboid cuboid) {
            return cuboid.drillDown(columns);
        }
    }

    public static class DrillUpOperation implements Operation {
        public DrillUpOperation() {
        }

        @Override
        public Cuboid get(Cuboid cuboid) {
            return cuboid.drillUp();
        }
    }

    public static class SliceOperation implements Operation {
        private final Object slice;

        public SliceOperation(Object slice) {
            this.slice = slice;
        }

        @Override
        public Cuboid get(Cuboid cuboid) {
            return cuboid.slice(slice);
        }
    }

    public static class DiceOperation implements Operation {
        private final Set<Object> dice;

        public DiceOperation(Set<Object> dice) {
            this.dice = dice;
        }

        @Override
        public Cuboid get(Cuboid cuboid) {
            return cuboid.dice(dice);
        }
    }

    public static class CollapseOperation implements Operation {
        public CollapseOperation() {
        }

        @Override
        public Cuboid get(Cuboid cuboid) {
            return cuboid.collapse();
        }
    }
}
