/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.aggregates.query;

import com.macrofocus.molap.aggregates.aggregation.AbstractAggregation;
import com.macrofocus.molap.aggregates.aggregation.Aggregation;
import com.macrofocus.molap.aggregates.cube.Dimensions;
import com.macrofocus.molap.aggregates.cube.Group;
import com.macrofocus.molap.aggregates.cuboid.Cuboid;

import java.util.*;

public class PivotQuery extends AbstractQuery {
    private final Query query;

    private Cuboid cuboid;
    private final Aggregation aggregation;
    private Aggregation[] aggregations;
    private Object[] columns;

    private List<Group> groups;

    final QueryListener queryListener = new QueryListener() {
        @Override
        public void queryChanged() {
            setDirty(true);
            notifyQueryChanged();
        }
    };

    public PivotQuery(Query query, Aggregation aggregation) {
        this.query = query;
        this.aggregation = aggregation;
        this.columns = null;

        query.addWeakQueryListener(queryListener);
    }

    public PivotQuery(Query query, Aggregation aggregation, Object[] columns) {
        this.query = query;
        this.aggregation = aggregation;
        this.columns = columns;

        query.addWeakQueryListener(queryListener);
    }

    private void run() {
        Cuboid cuboid = query.getCuboid().drillUp();
        List<Group> groups = new ArrayList<Group>();
        for (Group group : cuboid.getGroups()) {
            groups.add(group);
        }

        Object[] values;
        if(this.columns != null) {
            values = columns;
        } else {
            final Set<Object> v = query.getCuboid().getValues();
            values = v.toArray();
            Arrays.sort(values, new Comparator<Object>() {
                @Override
                public int compare(Object o1, Object o2) {
                    if (o1 == null) {
                        if (o2 == null) {
                            return 0;
                        } else {
                            return 1;
                        }
                    } else {
                        if (o2 == null) {
                            return -1;
                        } else {
                            return ((Comparable) o1).compareTo(o2);
                        }
                    }
                }
            });
        }

        Aggregation[] aggregations = new Aggregation[values.length];
        for (int i = 0; i < values.length; i++) {
            Object o = values[i];
            aggregations[i] = new PivotAggregation(o, aggregation);
        }

        this.cuboid = cuboid;
        this.groups = groups;
        this.aggregations = aggregations;

        setDirty(false);
    }

    @Override
    public Object getValue(Group path, Aggregation aggregation) {
        return aggregation.compute(query.getCuboid(), path);
    }

    @Override
    public Dimensions getDimensions() {
        return query.getDimensions().drillUp();
    }

    @Override
    public Aggregation[] getAggregations() {
        if(isDirty()) {
            run();
        }
        return aggregations;
    }

    @Override
    public List<Group> getGroups() {
        if(isDirty()) {
            run();
        }
        return groups;
    }

    @Override
    public Cuboid getCuboid() {
        if(isDirty()) {
            run();
        }
        return cuboid;
    }

    @Override
    public void setDice(Set<Object> valuesSets) {
        cuboid = cuboid.dice(valuesSets);
        setDirty(true);
        notifyQueryChanged();
    }

    private static class PivotAggregation extends AbstractAggregation {
        private final Object value;
        private final Aggregation aggregation;

        private PivotAggregation(Object value, Aggregation aggregation) {
            this.value = value;
            this.aggregation = aggregation;
        }

        @Override
        public Class getType() {
            return aggregation.getType();
        }

        @Override
        public Object compute(Cuboid cuboid, Group group) {
            Group g = group.drillDown(value);

            return aggregation.compute(cuboid, g);
        }

        @Override
        public String toString() {
            if(value != null) {
                return value.toString();
            } else {
                return null;
            }
        }
    }
}
