/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.aggregates.query;

import com.macrofocus.molap.aggregates.AggregateDataFrame;
import com.macrofocus.molap.aggregates.aggregation.Aggregation;
import com.macrofocus.molap.aggregates.cube.Group;
import com.macrofocus.molap.aggregates.cuboid.Cuboid;
import com.macrofocus.molap.dataframe.AbstractDataFrame;
import com.macrofocus.molap.dataframe.DataFrame;
import com.macrofocus.molap.dataframe.DataFrameEvent;
import com.macrofocus.molap.dataframe.ReIndexedDataFrame;
import com.macrofocus.molap.index.DefaultIndex;
import com.macrofocus.molap.index.Index;
import com.macrofocus.molap.index.MultiKey;
import com.macrofocus.molap.series.AbstractSeries;
import com.macrofocus.molap.series.Series;

public class QueryDataFrame<C> extends AbstractDataFrame<Group,String,Object> implements AggregateDataFrame<C> {
    private final Query query;
    private final boolean includeIndex;

    private Index<Group> rowIndex = null;
    private Index<String> columnIndex = null;

    private final QueryListener queryListener = new QueryListener() {
        @Override
        public void queryChanged() {
            rowIndex = null;
            columnIndex = null;

            notifyDataFrameChanged(new DataFrameEvent<Group, String>(null, null, true));
        }
    };

    public QueryDataFrame(Query query, boolean includeIndex) {
        this.query = query;
        this.includeIndex = includeIndex;

        query.addWeakQueryListener(queryListener);
    }

    @Override
    public Cuboid getCuboid() {
        return query.getCuboid();
    }

    @Override
    public AggregateDataFrame<C> drillDown(C... columns) {
        return new QueryDataFrame(query.drillDown(columns), includeIndex);
    }

    @Override
    public AggregateDataFrame<C> drillUp() {
        return new QueryDataFrame<C>(query.drillUp(), includeIndex);
    }

    @Override
    public AggregateDataFrame<C> order(Aggregation... aggregations) {
        return new QueryDataFrame<C>(query.order(aggregations), includeIndex);
    }

    @Override
    public AggregateDataFrame<C> pivot(Aggregation aggregation) {
        return new QueryDataFrame<C>(query.pivot(aggregation), includeIndex);
    }

    @Override
    public AggregateDataFrame<C> on(Cuboid... cuboids) {
        return new QueryDataFrame<C>(query.on(cuboids), includeIndex);
    }

    @Override
    public Class getRowClass(Group row) {
        return Object.class;
    }

    @Override
    public Class getColumnClass(String column) {
        Integer columnIndex = getColumnIndex().getAddress(column);

        if(columnIndex != null) {
            if (includeIndex) {
                if (columnIndex < query.getDimensions().path.length) {
                    return query.getCuboid().getCube().getDataFrame().getColumnClass(query.getDimensions().path[columnIndex]);
                }
                columnIndex -= query.getDimensions().path.length;
            }
            return query.getAggregations()[columnIndex].getType();
        } else {
            return Object.class;
        }
    }

    @Override
    public Object getValueAt(Group path, String column) {
        Integer columnIndex = getColumnIndex().getAddress(column);

        if(columnIndex != null) {
            assert path != null;

            if(includeIndex) {
                if (columnIndex < query.getDimensions().path.length) {
                    if (query.getDimensions() != null) {
                        final Object c = query.getDimensions().path[columnIndex];
                        columnIndex = query.getDimensions().getIndex(c);
                    }
                    return path.getPath(columnIndex);
                }
                columnIndex -= query.getDimensions().path.length;
            }
            return query.getValue(path, query.getAggregations()[columnIndex]);
        } else {
            return null;
        }
    }

    @Override
    public Series<String, ?> getRow(Group row) {
        return null;
    }

    @Override
    public Series<Group, Object> getColumn(String column) {
        int columnIndex = getColumnIndex().getAddress(column);
        if(includeIndex) {
            if (columnIndex < query.getDimensions().path.length) {
                return new DimensionSeries(query.getCuboid().getCube().getDataFrame(), column, columnIndex);
            }
            columnIndex -= query.getDimensions().path.length;
        }
        return new AggregationSeries(query.getAggregations()[columnIndex], column);
    }

    @Override
    public DataFrame<String, Group, ?> append(Series<String, ?> series) {
        return null;
    }

    @Override
    public Iterable<Group> rows() {
        return query.getGroups();
    }

    @Override
    public Iterable<String> columns() {
        return getColumnIndex().keys();
    }

    @Override
    public Group getRowKey(int index) {
        final Group path = query.getGroups().get(index);
        return path;
    }

    @Override
    public String getColumnKey(int index) {
        if(includeIndex) {
            if (index < query.getDimensions().path.length) {
                return query.getDimensions().path[index].toString();
            }
            index -= query.getDimensions().path.length;
        }
        return query.getAggregations()[index].getName();
    }

    @Override
    public int getRowAddress(Group row) {
        return getRowIndex().getAddress(row);
    }

    @Override
    public int getColumnAddress(String column) {
        return getColumnIndex().getAddress(column);
    }

    @Override
    public int getRowCount() {
        return query.getGroups().size();
    }

    @Override
    public int getColumnCount() {
        if(includeIndex) {
            return query.getDimensions().path.length + query.getAggregations().length;
        } else {
            return query.getAggregations().length;
        }
    }

    @Override
    public DataFrame<MultiKey, String, Object> reindexRows(final String... columns) {
        return new ReIndexedDataFrame<MultiKey, String, Object, Group, String>(this, new ReIndexedDataFrame.ReIndexRecipe<MultiKey, String>() {
            @Override
            public Index<MultiKey> builRowdIndex() {
                int r = 0;
                final MultiKey[] keys = new MultiKey[getRowCount()];
                for(Group row: rows()) {
                    final Object[] values = new Object[columns.length];
                    int c = 0;
                    for (String column : columns) {
                        values[c++] = getValueAt(row, column);
                    }

                    MultiKey key = new MultiKey(values);
                    keys[r++] = key;
                }

                final DefaultIndex<MultiKey> index = new DefaultIndex<MultiKey>(keys);
                return index;
            }

            @Override
            public Index<String> buildColumnIndex() {
                return getColumnIndex();
            }
        });
    }

    @Override
    public DataFrame join(Series series, String[] columns) {
        return null;
    }

    public Index<Group> getRowIndex() {
        if(rowIndex == null) {
            Group[] groups = new Group[query.getGroups().size()];
            int i = 0;
            for(Group g: query.getGroups()) {
                groups[i++] = g;
            }

            rowIndex = new DefaultIndex<Group>(groups);
        }
        return rowIndex;
    }

    public Index<String> getColumnIndex() {
        if(columnIndex == null) {
            String[] names = new String[getColumnCount()];
            for(int c = 0; c < getColumnCount(); c++) {
                names[c] = getColumnKey(c);
            }

            columnIndex = new DefaultIndex<String>(names);
        }

        return columnIndex;
    }

    private class DimensionSeries extends AbstractSeries<Group,Object> {
        private final DataFrame<Group,String,Object> dataFrame;
        private final String column;
        private final int columnIndex;

        private DimensionSeries(DataFrame<Group, String, Object> dataFrame, String column, int columnIndex) {
            this.dataFrame = dataFrame;
            this.column = column;
            this.columnIndex = columnIndex;
        }

        @Override
        public Object getName() {
            return column;
        }

        @Override
        public Class getType() {
            return dataFrame.getColumnClass(column);
        }

        @Override
        public Object get(Group key) {
            return key.getPath(columnIndex);
        }

        @Override
        public Group getKey(int i) {
            return getRowKey(i);
        }

        @Override
        public Iterable<Group> keys() {
            return rows();
        }

        @Override
        public <L> Series<L, Object> reindex(L... keys) {
            return null;
        }

        @Override
        public Series<Group, Object> head(int count) {
            return null;
        }

        @Override
        public Series<Group, Object> tail(int count) {
            return null;
        }
    }

    private class AggregationSeries extends AbstractSeries<Group,Object> {
        private final Aggregation aggregation;
        private final String column;

        private AggregationSeries(Aggregation aggregation, String column) {
            this.aggregation = aggregation;
            this.column = column;
        }

        @Override
        public Object getName() {
            return column;
        }

        @Override
        public Class getType() {
            return Double.class;
        }

        @Override
        public Object get(Group key) {
            return query.getValue(key, aggregation);
        }

        @Override
        public Group getKey(int i) {
            return getRowKey(i);
        }

        @Override
        public Iterable<Group> keys() {
            return rows();
        }

        @Override
        public <L> Series<L, Object> reindex(L... keys) {
            return null;
        }

        @Override
        public Series<Group, Object> head(int count) {
            return null;
        }

        @Override
        public Series<Group, Object> tail(int count) {
            return null;
        }
    }
}
