/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.dataframe;

import com.macrofocus.colormap.SimpleColorMap;
import com.macrofocus.filter.MutableFilter;
import com.macrofocus.molap.aggregates.AggregateDataFrame;
import com.macrofocus.molap.aggregates.aggregation.*;
import com.macrofocus.molap.aggregates.cube.Cube;
import com.macrofocus.molap.aggregates.cube.InMemoryCube;
import com.macrofocus.molap.aggregates.query.CuboidQuery;
import com.macrofocus.molap.aggregates.query.QueryDataFrame;
import com.macrofocus.molap.dataframe.swing.DataFrameTableModel;
import com.macrofocus.molap.index.DefaultIndex;
import com.macrofocus.molap.index.Index;
import com.macrofocus.molap.index.MultiKey;
import com.macrofocus.molap.series.AbstractSeries;
import com.macrofocus.molap.series.Series;

import javax.swing.table.TableModel;
import java.io.PrintStream;
import java.lang.ref.WeakReference;
import java.text.DecimalFormat;
import java.util.Comparator;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

/**
 * This class provides a skeletal implementation of the DataFrame interface to minimize the effort required to
 * implement this interface.
 */
public abstract class AbstractDataFrame<R, C, V> implements DataFrame<R, C, V> {
    private final List<DataFrameListener<R, C>> listeners;

    private Cube cube = null;

    private static final Comparator universalComparator = new SimpleColorMap.UniversalComparator();

    protected AbstractDataFrame() {
        listeners = new CopyOnWriteArrayList<DataFrameListener<R, C>>();
    }

    @Override
    public Iterable<R> rows() {
        return getRowIndex().keys();
    }

    @Override
    public Iterable<C> columns() {
        return getColumnIndex().keys();
    }

    @Override
    public R getRowKey(int index) {
        return getRowIndex().getKey(index);
    }

    @Override
    public C getColumnKey(int index) {
        return getColumnIndex().getKey(index);
    }

    @Override
    public int getRowAddress(R r) {
        return getRowIndex().getAddress(r);
    }

    @Override
    public int getColumnAddress(C c) {
        return getColumnIndex().getAddress(c);
    }

    @Override
    public int getRowCount() {
        return getRowIndex().getSize();
    }

    @Override
    public int getColumnCount() {
        return getColumnIndex().getSize();
    }

    @Override
    public Series<R, V> getColumn(C c) {
        return new ColumnSeries(c);
    }

    @Override
    public DataFrame<R, C, V> orderRows(final C... columns) {
        return new ReIndexedDataFrame<>(this, new ReIndexedDataFrame.ReIndexRecipe<R, C>() {
            @Override
            public Index<R> builRowdIndex() {
                return new DefaultIndex<>(AbstractDataFrame.this.getRowIndex(), new Comparator<R>() {
                    @Override
                    public int compare(R o1, R o2) {
                        for (C c : columns) {
                            Object v1 = AbstractDataFrame.this.getValueAt(o1, c);
                            Object v2 = AbstractDataFrame.this.getValueAt(o2, c);

                            final int compare = universalComparator.compare(v1, v2);
                            if (compare != 0) {
                                return compare;
                            }
                        }
                        final int x = AbstractDataFrame.this.getRowAddress(o1);
                        final int y = AbstractDataFrame.this.getRowAddress(o2);
                        return (x < y) ? -1 : ((x == y) ? 0 : 1);
                    }
                });
            }

            @Override
            public Index<C> buildColumnIndex() {
                return getColumnIndex();
            }
        });
    }

    @Override
    public FilterDataFrame filter(MutableFilter<R> filter) {
        return new FilterDataFrame<R,C,V>(this, filter);
    }

    protected Cube getCube() {
        if(cube == null) {
            cube = new InMemoryCube(this);
        }
        return cube;
    }

    @Override
    public AggregateDataFrame<C> aggregate(Aggregation... aggregations) {
        return new QueryDataFrame(new CuboidQuery(getCube().getApexCuboid(), aggregations), true);
    }

    @Override
    public SumAggregation getSum(C column) {
        return new SumAggregation(getColumn(column));
    }

    @Override
    public CountAggregation getCount(C column) {
        return new CountAggregation(getColumn(column));
    }

    @Override
    public MinAggregation getMin(C column) {
        return new MinAggregation(getColumn(column));
    }

    @Override
    public MaxAggregation getMax(C column) {
        return new MaxAggregation(getColumn(column));
    }

    @Override
    public MeanAggregation getMean(C column) {
        return new MeanAggregation(getColumn(column));
    }

    @Override
    public VarianceAggregation getVariance(C column) {
        return new VarianceAggregation(getColumn(column));
    }

    @Override
    public VarianceAggregation getVariance(C column, C population) {
        return new VarianceAggregation(getColumn(column), getColumn(population));
    }

    @Override
    public StdDevAggregation getStdDev(C column) {
        return new StdDevAggregation(getColumn(column));
    }

    @Override
    public CountDistinctAggregation getCountDistinct(C column) {
        return new CountDistinctAggregation(getColumn(column));
    }

    @Override
    public String getColumnName(C column) {
        if(column != null) {
            return column.toString();
        } else {
            return null;
        }
    }

    @Override
    public void print() {
        PrintStream out = System.out;

        for(C column: columns()) {
            out.printf("%20s", abbreviate(getColumnName(column), 19));

        }
        out.println();

        for(C column: columns()) {
            Class value = getColumnClass(column);
            out.printf("%20s", abbreviate(value != null ? value.getSimpleName() : "", 19));
        }
        out.println();

        for(R row: rows()) {
//            System.out.print(row + "\t");
            for(C column: columns()) {
                Object value = getValueAt(row, column);
                if(value != null) {
                    if(value instanceof Number) {
                        out.printf("%20s", abbreviate(DecimalFormat.getNumberInstance().format(value), 19));
                    } else {
                        out.printf("%20s", abbreviate(value.toString(), 19));
                    }
                } else {
                    out.printf("%20s", "", 19);

                }
            }
            out.println();
        }
    }

    @Override
    public void print(PrintStream out, String caption, boolean html) {
        TableModel tableModel = new DataFrameTableModel(this);

        if(!html) {
            if(caption != null) {
                System.out.println(caption);
            }

            int cols = tableModel.getColumnCount();
            for (int i = 0; i < cols; i++) {
                final String o = tableModel.getColumnName(i);
                final String s;
                if(o != null) {
                    s = o;
                } else {
                    s = "n.a.";
                }
                out.printf("%20s", abbreviate(s, 19));
            }
            //        System.out.println();
            //        for (int i = 0; i < cols; i++) {
            //            final String o = tableModel.getColumnClass(i).getSimpleName();
            //            final String s;
            //            if(o != null) {
            //                s = o;
            //            } else {
            //                s = "n.a.";
            //            }
            //            out.printf("%20s", abbreviate(s, 19));
            //        }
            out.println();
            for(int row = 0; row < tableModel.getRowCount(); row++) {
                for (int i = 0; i < cols; i++) {
                    final Object o = tableModel.getValueAt(row, i);
                    final String s;
                    if(o != null) {
                        s = o.toString();
                    } else {
                        s = "n.a.";
                    }
                    out.printf("%20s", abbreviate(s, 19));
                }
                out.println();
            }
        } else {
            out.println("<table style=\"width: auto; border-collapse: separate;\">");

            if(caption != null) {
                out.println("<caption>" + caption + "</caption>");
            }

            out.println("<tbody>");
            out.println("<tr>");
            int cols = tableModel.getColumnCount();
            for (int i = 0; i < cols; i++) {
                final String o = tableModel.getColumnName(i);
                final String s;
                if(o != null) {
                    s = o;
                } else {
                    s = "n.a.";
                }
                out.print("<th>");
                out.print(s);
                out.println("</th>");
            }
            out.println("</tr>");

            for(int row = 0; row < tableModel.getRowCount(); row++) {
                out.println("<tr>");
                for (int i = 0; i < cols; i++) {
                    final Object o = tableModel.getValueAt(row, i);
                    final String s;
                    if(o != null) {
                        s = o.toString();
                    } else {
                        s = "n.a.";
                    }
                    if(o instanceof Number) {
                        out.print("<td align=\"right\">");
                    } else {
                        out.print("<td>");
                    }
                    out.print(s);
                    out.println("</td>");
                }
                out.println("</tr>");
            }
            out.println("</tbody>");

            out.println("</table>");
        }
    }

    private static String abbreviate(String str, int maxWidth) {
        if(str == null) {
            return null;
        }
        if (str.length() <= maxWidth) {
            return str;
        }
        return str.substring(0, maxWidth - 3) + "...";
    }

    @Override
    public void benchmark() {
        long now = System.currentTimeMillis();

        for(int i = 0; i < 100; i++) {
//            for (K column : columns()) {
//            }
//
//            for (K column : columns()) {
//            }

            for (C column : columns()) {
                Series series = getColumn(column);
                for (R row : rows()) {
//                    Object value = getValue(row, column);
                    Object value = series.get(row);
                }
            }
        }

        System.out.println("Benchmark: " + (System.currentTimeMillis() - now) + " ms");
    }

    @Override
    public DataFrame<MultiKey, C, V> reindexRows(final C... columns) {
        return new ReIndexedDataFrame<MultiKey, C, V, R, C>(this, new ReIndexedDataFrame.ReIndexRecipe<MultiKey, C>() {
            @Override
            public Index<MultiKey> builRowdIndex() {
                int r = 0;
                final MultiKey[] keys = new MultiKey[getRowCount()];
                for(R row: rows()) {
                    final Object[] values = new Object[columns.length];
                    int c = 0;
                    for (C column : columns) {
                        values[c++] = getValueAt(row, column);
                    }

                    MultiKey key = new MultiKey(values);
                    keys[r++] = key;
                }

                final DefaultIndex<MultiKey> index = new DefaultIndex<MultiKey>(keys);
                return index;
            }

            @Override
            public Index<C> buildColumnIndex() {
                return getColumnIndex();
            }
        });
    }

    public void addDataFrameListener(final DataFrameListener<R, C> listener) {
        synchronized (listeners) {
            listeners.add(listener);
        }
    }

    public void addWeakDataFrameListener(final DataFrameListener<R, C> listener) {
        synchronized (listeners) {
            listeners.add(new WeakDataFrameListener<R, C>(listener));
        }
    }

    public void removeDataFrameListener(final DataFrameListener<R, C> listener) {
        synchronized (listeners) {
            if (listener instanceof WeakDataFrameListener) {
                boolean removed = listeners.remove(listener);
                assert removed : listener;
            } else {
                DataFrameListener<R, C> toRemove = null;

                for (DataFrameListener<R, C> dataFrameListener : listeners) {
                    DataFrameListener comparable;
                    if (dataFrameListener instanceof WeakDataFrameListener) {
                        comparable = ((WeakDataFrameListener) dataFrameListener).getReference();
                    } else {
                        comparable = dataFrameListener;
                    }

                    if (listener.equals(comparable)) {
                        toRemove = dataFrameListener;
                    }
                }

                if (toRemove != null) {
                    boolean removed = listeners.remove(toRemove);
                    assert removed : listener;
                }
            }
        }
    }

    public void removeDataFrameListeners() {
        synchronized (listeners) {
            listeners.clear();
        }
    }

    /**
     * Fires an event event to listeners that have been registered to track changes to the data frame.
     *
     * @param event the event describing the changes
     */
    protected void notifyDataFrameChanged(DataFrameEvent<R, C> event) {
        for (final DataFrameListener<R, C> listener : listeners) {
            listener.dataFrameChanged(event);
        }
    }

    private class WeakDataFrameListener<R,K> implements DataFrameListener<R,K> {
        private final WeakReference<DataFrameListener<R,K>> l_ref;

        public WeakDataFrameListener(final DataFrameListener<R,K> listener) {
            this.l_ref = new WeakReference<DataFrameListener<R,K>>(listener);
        }

        public void dataFrameChanged(DataFrameEvent event) {
            final DataFrameListener<R,K> l = getReference();
            if (l != null) {
                l.dataFrameChanged(event);
            } else {
                removeDataFrameListener((DataFrameListener) this);
            }
        }

        public DataFrameListener<R,K> getReference() {
            return l_ref.get();
        }
    }

    public class ColumnSeries extends AbstractSeries<R,V> {
        private C column;

        private ColumnSeries(C column) {
            assert column != null;

            this.column = column;
        }

        @Override
        public Object getName() {
            return column;
        }

        @Override
        public Class getType() {
            return getColumnClass(column);
        }

        @Override
        public V get(R key) {
            return getValueAt(key, column);
        }

        @Override
        public R getKey(int i) {
            return getRowIndex().getKey(i);
        }

        @Override
        public Iterable<R> keys() {
            return getRowIndex().keys();
        }

        @Override
        public <L> Series<L, V> reindex(L... keys) {
            return null;
        }

        @Override
        public Series<R, V> head(int count) {
            return null;
        }

        @Override
        public Series<R, V> tail(int count) {
            return null;
        }
    }
}
