/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.dataframe;

import com.macrofocus.filter.MutableFilter;
import com.macrofocus.molap.aggregates.AggregateDataFrame;
import com.macrofocus.molap.aggregates.aggregation.*;
import com.macrofocus.molap.index.Index;
import com.macrofocus.molap.index.MultiKey;
import com.macrofocus.molap.series.Series;

import java.io.PrintStream;

/**
 * An indexed table structure, i.e. a 2D array of values that can be queried row and/or column-wise using labels.
 * This is similar to the DataFrame that can be found in the R and Pandas libraries.
 *
 * @param <Row> the type of row keys maintained used to query this data frame.
 * @param <Column> the type of column keys maintained used to query this data frame.
 * @param <V> the type of values
 */
public interface DataFrame<Row, Column,V> {
    /**
     * Returns the name of the column. This is a convenience method for labeling purpose.
     *
     * @param column the key of the column
     * @return the name of the column
     */
    String getColumnName(Column column);

    /**
     * Returns the most specific superclass for all cell values in a row.
     *
     * @param row the key of the row
     * @return the common ancestor class of the object values in the row.
     */
    Class getRowClass(Row row);

    /**
     * Returns the most specific superclass for all cell values in a column
     *
     * @param column the key of the column
     * @return the common ancestor class of the object values in the column.
     */
    Class getColumnClass(Column column);

    /**
     * Returns the value for the cell at the intersection of the
     * <code>column</code> key and <code>row</code> key.
     *
     * @param   row        the row key whose value is to be queried
     * @param   column     the column key whose value is to be queried
     * @return  the value Object at the specified cell
     */
    V getValueAt(Row row, Column column);

    /**
     * Returns a series of all the values of a given row.
     *
     * @param row the row key
     * @return a Series object
     */
    Series<Column,?> getRow(Row row);

    /**
     * Returns a series of all the values of a given column.
     *
     * @param column the row key
     * @return a Series object
     */
    Series<Row,V> getColumn(Column column);

    /**
     * Returns the row keys.
     *
     * @return the row keys
     */
    Iterable<Row> rows();

    /**
     * Returns the column keys.
     *
     * @return the column keys
     */
    Iterable<Column> columns();

    /**
     * Returns the row key at the specified absolute index. This is the inverse of {@link #getRowAddress(Object)}.
     *
     * @param index the index
     * @return the row key
     */
    Row getRowKey(int index);

    /**
     * Returns the column key at the specified absolute index. This is the inverse of {@link #getColumnAddress(Object)}.
     *
     * @param index the index
     * @return the column key
     */
    Column getColumnKey(int index);

    /**
     * Returns the absolute index for the specified row key. This is the inverse of {@link #getRowKey(int)}.
     *
     * @param row the row key
     * @return the absolute index of the specified key.
     */
    int getRowAddress(Row row);

    /**
     * Returns the absolute index for the specified column key. This is the inverse of {@link #getColumnKey(int)}.
     *
     * @param column the column key
     * @return the absolute index of the specified key.
     */
    int getColumnAddress(Column column);

    /**
     * Returns the number of rows contained by the this data frame.
     *
     * @return the number of rows.
     */
    int getRowCount();

    /**
     * Returns the number of columns contained by this data frame.
     *
     * @return the number of columns.
     */
    int getColumnCount();

    /**
     * Returns a new data frame reindexed using the values coming from the specified columns.
     *
     * @param columns the columns to use for the label values
     * @return the reindexed data frame.
     */
    DataFrame<MultiKey, Column,V> reindexRows(Column... columns);

    /**
     * Gets the index used to access the rows.
     *
     * @return the row index
     */
    Index<Row> getRowIndex();

    /**
     * Gets the index used to access the columns.
     *
     * @return the column index
     */
    Index<Column> getColumnIndex();

    /**
     * Returns a new data frame reordered using the values coming from the specified columns.
     *
     * @param columns the columns to use for sorting
     * @return the sorted data frame.
     */
    DataFrame<Row, Column, V> orderRows(Column... columns);

    /**
     * Returns a new data frame those rows are filtered by the specified filter model.
     *
     * @param filter the filter model
     * @return a filtered data frame
     */
    FilterDataFrame<Row, Column, V> filter(MutableFilter<Row> filter);

    /**
     * Returns a new data frame suitable for data aggregation. The additional methods provided to customize the
     * aggregation criteria are provided by the {@link com.macrofocus.molap.aggregates.AggregateDataFrame}.
     *
     * @param aggregation the aggregation methods to use as columns
     * @return an aggregated data frame.
     */
    AggregateDataFrame<Column> aggregate(Aggregation... aggregation);

    DataFrame<Column, Row,?> append(Series<Column,?> series);

    DataFrame join(Series series, Column[] columns);

    /**
     * Returns the aggregation method for computing the sum.
     *
     * @param column the column key
     * @return the aggregation method
     */
    SumAggregation getSum(Column column);

    /**
     * Returns the aggregation method for counting the number of values.
     *
     * @param column the column key
     * @return the aggregation method
     */
    CountAggregation getCount(Column column);

    /**
     * Returns the aggregation method for finding the minimum value.
     *
     * @param column the column key
     * @return the aggregation method
     */
    MinAggregation getMin(Column column);

    /**
     * Returns the aggregation method for finding the maximum value.
     *
     * @param column the column key
     * @return the aggregation method
     */
    MaxAggregation getMax(Column column);

    /**
     * Returns the aggregation method for computing the mean value.
     *
     * @param column the column key
     * @return the aggregation method
     */
    MeanAggregation getMean(Column column);

    /**
     * Returns the aggregation method for computing the variance.
     *
     * @param column the column key
     * @return the aggregation method
     */
    VarianceAggregation getVariance(Column column);

    /**
     * Returns the aggregation method for computing the variance by factoring in the specified population.
     *
     * @param column the column key
     * @param population the key of the column for the population
     * @return the aggregation method
     */
    VarianceAggregation getVariance(Column column, Column population);

    /**
     * Returns the aggregation method for computing the standard deviation.
     *
     * @param column the column key
     * @return the aggregation method
     */
    StdDevAggregation getStdDev(Column column);

    /**
     * Returns the aggregation method for counting the number of distinct values.
     *
     * @param column the column key
     * @return the aggregation method
     */
    CountDistinctAggregation getCountDistinct(Column column);

    /**
     * Display the content of the data frame to the console.
     */
    void print();

    /**
     * Display the content of the data frame to the specified output.
     *
     * @param out the output
     * @param caption the title
     * @param html false for text output, true for HTML formatting
     */
    void print(PrintStream out, String caption, boolean html);

    /**
     * Perform some repetitive operation to estimate the performance of the data frame.
     */
    void benchmark();

    /**
     * Add a listener to the list that's notified each time a change to the data frame occurs.
     *
     * @param listener the DataFrameListener
     */
    void addDataFrameListener(DataFrameListener<Row, Column> listener);

    /**
     * Add a listener to the list that's notified each time a change to the data frame occurs. The listener will
     * automatically be disposed of should no other object have a reference to it.
     *
     * @param listener the DataFrameListener
     */
    void addWeakDataFrameListener(DataFrameListener<Row, Column> listener);

    /**
     * Remove a listener to the list that's notified each time a change to the data frame occurs.
     *
     * @param listener the DataFrameListener
     */
    void removeDataFrameListener(DataFrameListener<Row, Column> listener);

    /**
     * Remove all listeners to the list that's notified each time a change to the data frame occurs.
     */
    void removeDataFrameListeners();
}
