/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.dataframe;

import com.macrofocus.data.db.MySQLDatabaseDriver;
import com.macrofocus.data.db.OracleDatabaseDriver;
import com.macrofocus.data.db.PostgreSQLDatabaseDriver;
import com.macrofocus.data.db.SQLServerDatabaseDriver;
import com.macrofocus.data.source.*;
import com.macrofocus.molap.dataframe.swing.SwingDataFrameFactory;
import com.macrofocus.molap.series.Series;

import java.io.IOException;
import java.net.URL;
import java.sql.SQLException;

/**
 * Factory for supporting the most common ways of instantiating <code>DataFrame</code> object.
 */
public class DataFrameFactory {
    private static final DataFrameFactory ourInstance = new DataFrameFactory();

    public static DataFrameFactory getInstance() {
        return ourInstance;
    }

    public static SwingDataFrameFactory getSwingInstance() {
        return new SwingDataFrameFactory();
    }

    protected DataFrameFactory() {
    }

    /**
     * Creates a data frame using a combination of series.
     *
     * @param series the series to combine
     * @param <R> the type of row keys maintained used to query this data frame.
     * @param <V> the type of values
     * @return a newly created data frame.
     */
    public static <R,V> DataFrame<R,Object,V> fromColumnSeries(Series<R, V>... series) {
        return new IndexedDataFrame<R,V>(series);
    }

    /**
     * Creates a data frame using a row-major two dimensional array.
     *
     * @param names the column names
     * @param classes the type of values hold by each column
     * @param data the data array
     * @param <V>  the type of values
     * @return a newly created data frame.
     */
    public static <V> MutableDataFrame<Integer,String,V> fromRowMajorArray(String[] names, Class[] classes, V[][] data) {
        return new DefaultDataFrame<V>(names, classes, data);
    }

    /**
     * Creates a data frame by loading a Comma-Separated Values (CSV) encoded file.
     * The macrofocus-data.jar library is required to use this method.
     *
     * @param url the URL of the file
     * @return a newly created data frame
     * @throws IOException
     */
    public static MutableDataFrame<Integer,String,Object> fromCSV(URL url) throws IOException {
        try {
            return new DefaultDataFrame<Object>(new CommaDelimitedDataSource(url).load(null));
        } catch (SQLException e) {
            e.printStackTrace();
            return null;
        }
    }

    /**
     * Creates a data frame by loading a tab-delimited encoded file.
     * The macrofocus-data.jar library is required to use this method.
     *
     * @param url the URL of the file
     * @return a newly created data frame
     * @throws IOException
     */
    public static MutableDataFrame<Integer,String,Object> fromTSV(URL url) throws IOException {
        try {
            return new DefaultDataFrame<Object>(new TabDelimitedDataSource(url).load(null));
        } catch (SQLException e) {
            e.printStackTrace();
            return null;
        }
    }

    /**
     * Creates a data frame by loading an Excel (.xls, .xlsx) file.
     * The macrofocus-data.jar and POI libraries are required to use this method.
     *
     * @param url the URL of the file
     * @return a newly created data frame
     * @throws IOException
     */
    public static MutableDataFrame<Integer,String,Object> fromExcel(URL url) throws IOException {
        try {
            return new DefaultDataFrame<Object>(new ExcelDataSource(url).load(null));
        } catch (SQLException e) {
            e.printStackTrace();
            return null;
        }
    }

    /**
     * Creates a data frame by loading a JSON encoded file.
     * The macrofocus-data.jar library is required to use this method.
     *
     * @param url the URL of the file
     * @return a newly created data frame
     * @throws IOException
     */
    public static MutableDataFrame<Integer,String,Object> fromJSON(URL url) throws IOException {
        try {
            return new DefaultDataFrame<Object>(new JSONDataSource(url).load(null));
        } catch (SQLException e) {
            e.printStackTrace();
            return null;
        }
    }

    /**
     * Creates a data frame by retrieving the result of a query made to a MySQL database.
     * The macrofocus-data.jar library is required to use this method.
     *
     * @param url the JDBC URL, e.g. "jdbc:mysql://mysql.example.com:3306/mydatabase"
     * @param username the username to authenticate the connection
     * @param password the password to authenticate the connection
     * @param query the SQL query to submit, e.g. "SELECT * FROM mytable";
     * @return a newly created data frame
     * @throws IOException
     * @throws SQLException
     */
    public static DataFrame<Integer,String,Object> fromMySQL(String url, String username, String password, String query) throws IOException,SQLException {
        return new DefaultDataFrame<Object>(new JDBCDataSource(new MySQLDatabaseDriver(), url, username, password, query).load(null));
    }

    /**
     * Creates a data frame by retrieving the result of a query made to a PostgreSQL database.
     * The macrofocus-data.jar library is required to use this method.
     *
     * @param url the JDBC URL, e.g. "jdbc:postgresql://mysql.example.com/mydatabase"
     * @param username the username to authenticate the connection
     * @param password the password to authenticate the connection
     * @param query the SQL query to submit, e.g. "SELECT * FROM mytable";
     * @return a newly created data frame
     * @throws IOException
     * @throws SQLException
     */
    public static DataFrame<Integer,String,Object> fromPostgreSQL(String url, String username, String password, String query) throws IOException,SQLException {
        return new DefaultDataFrame<Object>(new JDBCDataSource(new PostgreSQLDatabaseDriver(), url, username, password, query).load(null));
    }

    /**
     * Creates a data frame by retrieving the result of a query made to an Oracle database.
     * The macrofocus-data.jar library is required to use this method.
     *
     * @param url the JDBC URL, e.g. "jdbc:oracle:thin://oracle.example.com/mydatabase"
     * @param username the username to authenticate the connection
     * @param password the password to authenticate the connection
     * @param query the SQL query to submit, e.g. "SELECT * FROM mytable";
     * @return a newly created data frame
     * @throws IOException
     * @throws SQLException
     */
    public static DataFrame<Integer,String,Object> fromOracle(String url, String username, String password, String query) throws IOException,SQLException {
        return new DefaultDataFrame<Object>(new JDBCDataSource(new OracleDatabaseDriver(), url, username, password, query).load(null));
    }

    /**
     * Creates a data frame by retrieving the result of a query made to a Microsoft SQL Server database.
     * The macrofocus-data.jar library is required to use this method.
     *
     * @param url the JDBC URL, e.g. "jdbc:sqlserver://sqlserver.example.com/mydatabase"
     * @param username the username to authenticate the connection
     * @param password the password to authenticate the connection
     * @param query the SQL query to submit, e.g. "SELECT * FROM mytable";
     * @return a newly created data frame
     * @throws IOException
     * @throws SQLException
     */
    public static DataFrame<Integer,String,Object> fromSQLServer(String url, String username, String password, String query) throws IOException,SQLException {
        return new DefaultDataFrame<Object>(new JDBCDataSource(new SQLServerDatabaseDriver(), url, username, password, query).load(null));
    }
}
