/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.dataframe;

import com.macrofocus.molap.aggregates.cuboid.Cuboid;
import com.macrofocus.molap.index.DefaultIndex;
import com.macrofocus.molap.index.Index;
import com.macrofocus.molap.index.IntegerIndex;
import com.macrofocus.molap.index.MultiKey;
import com.macrofocus.molap.series.AbstractSeries;
import com.macrofocus.molap.series.MutableSeries;
import com.macrofocus.molap.series.Series;
import com.macrofocus.molap.series.SeriesFactory;

import javax.swing.table.TableModel;
import java.lang.reflect.Array;
import java.util.HashSet;
import java.util.Set;

/**
 * A data frame with direct integer indexing on the rows and textual indexing on the columns.
 *
 * @param <V> the type of values
 */
public class DefaultDataFrame<V> extends AbstractDataFrame<Integer,String,V> implements MutableDataFrame<Integer,String,V> {
    private final MutableSeries<Integer,V>[] series;
    private final DefaultIndex<String> columnIndex;
    private final Index<Integer> rowIndex;

    public DefaultDataFrame(TableModel tableModel) {
        series = (MutableSeries<Integer,V>[])new MutableSeries[tableModel.getColumnCount()];
        for(int c = 0; c < tableModel.getColumnCount(); c++) {
            V[] values = (V[])Array.newInstance(tableModel.getColumnClass(c), tableModel.getRowCount());
            for(int r = 0; r < tableModel.getRowCount(); r++) {
                final V valueAt = (V)tableModel.getValueAt(r, c);
                try {
                    if(valueAt != Cuboid.ALL) {
                        try {
                            values[r] = valueAt;
                        } catch (ArrayStoreException e) {
                            e.printStackTrace();
                        }
                    } else {
                        values[r] = null;
                    }
                } catch (ArrayStoreException e) {
                     System.err.println(valueAt + ", " + valueAt.getClass() + ": " + tableModel.getColumnClass(c));
                    e.printStackTrace();
                }
            }
            series[c] = SeriesFactory.<V>create(values);
        }

        String[] columns = new String[tableModel.getColumnCount()];
        Set<String> unique = new HashSet<>();
        for (int i = 0; i < tableModel.getColumnCount(); i++) {
            if(tableModel.getColumnName(i) != null) {
                String header = tableModel.getColumnName(i).intern();
                int c = 0;
                while(unique.contains(header)) {
                    header = header + "-" + c;
                }
                columns[i] = header;
                unique.add(header);
            } else {
                columns[i] = Integer.toString(i);
            }
        }
        columnIndex = new DefaultIndex<String>(columns);

        rowIndex = new IntegerIndex(0, tableModel.getRowCount() - 1);
    }

    public DefaultDataFrame(String[] names, Class[] classes, V[][] data) {
        series = (MutableSeries<Integer,V>[])new MutableSeries[names.length];
        int rowCount = data.length;
        for(int c = 0; c < names.length; c++) {
            V[] values = (V[])Array.newInstance(classes[c], rowCount);
            for(int r = 0; r < rowCount; r++) {
                values[r] = (V)data[r][c];
            }
            series[c] = SeriesFactory.<V>create(names[c].intern(), values);
        }

        String[] columns = new String[names.length];

        for (int i = 0; i < columns.length; i++) {
            columns[i] = names[i].intern();
        }

        System.arraycopy(names, 0, columns, 0, names.length);
        columnIndex = new DefaultIndex<String>(columns);

        rowIndex = new IntegerIndex(0, rowCount - 1);
    }

    public DefaultDataFrame(DataFrame dataFrame) {
        series = (MutableSeries<Integer,V>[])new MutableSeries[dataFrame.getColumnCount()];
        int c = 0;
        for(Object column: dataFrame.columns()) {
            V[] values = (V[])Array.newInstance(dataFrame.getColumnClass(column), dataFrame.getRowCount());
            Series s = dataFrame.getColumn(column);
            int r = 0;
            for(Object row: dataFrame.rows()) {
                values[r] = (V)s.get(row);
                r++;
            }
            series[c] = SeriesFactory.<V>create(values);
            c++;
        }

        String[] columns = new String[dataFrame.getColumnCount()];
        for (int i = 0; i < dataFrame.getColumnCount(); i++) {
            columns[i] = dataFrame.getColumnKey(i).toString().intern();
        }
        columnIndex = new DefaultIndex<String>(columns);

        rowIndex = new IntegerIndex(0, dataFrame.getRowCount() - 1);
    }

    public DefaultDataFrame(Index<Integer> rowIndex, DefaultIndex<String> columnIndex, MutableSeries<Integer, V>[] series) {
        this.rowIndex = rowIndex;
        this.columnIndex = columnIndex;
        this.series = series;
    }

    public DefaultIndex<String> getColumnIndex() {
        return columnIndex;
    }

    public Index<Integer> getRowIndex() {
        return rowIndex;
    }

    @Override
    public Iterable<String> columns() {
        return columnIndex.keys();
    }

    @Override
    public Class getRowClass(Integer row) {
        return Object.class;
    }

    @Override
    public Class getColumnClass(String column) {
        assert columnIndex.getAddress(column) != null: "Could not find " + column;

        return series[columnIndex.getAddress(column)].getType();
    }

    @Override
    public Series<String, ?> getRow(Integer row) {
        return null;
    }

    @Override
    public Series<Integer, V> getColumn(String column) {
        return new ColumnSeries2(column);
    }

    @Override
    public V getValueAt(Integer row, String column) {
        return series[columnIndex.getAddress(column)].get(rowIndex.getAddress(row));
    }

    @Override
    public void setValueAt(Integer row, String column, V value) {
        series[columnIndex.getAddress(column)].set(row, value);
    }

    @Override
    public DataFrame<String, Integer, ?> append(Series<String, ?> series) {
        return null;
    }

    @Override
    public Iterable<Integer> rows() {
        return rowIndex.keys();
    }

    @Override
    public Integer getRowKey(int index) {
        return rowIndex.getKey(index);
    }

    @Override
    public String getColumnKey(int index) {
        return columnIndex.getKey(index);
    }

    @Override
    public int getRowAddress(Integer row) {
        return rowIndex.getAddress(row);
    }

    @Override
    public int getColumnAddress(String column) {
        return columnIndex.getAddress(column);
    }

    @Override
    public int getRowCount() {
        return rowIndex.getSize();
    }

    @Override
    public int getColumnCount() {
        return columnIndex.getSize();
    }

    @Override
    public DataFrame<MultiKey,String,V> reindexRows(final String... columns) {
        return new ReIndexedDataFrame<MultiKey, String, V, Integer, String>(this, new ReIndexedDataFrame.ReIndexRecipe<MultiKey, String>() {
            @Override
            public Index<MultiKey> builRowdIndex() {
                int r = 0;
                final MultiKey[] keys = new MultiKey[getRowCount()];
                for(Integer row: rows()) {
                    final Object[] values = new Object[columns.length];
                    int c = 0;
                    for (String column : columns) {
                        values[c++] = getValueAt(row, column);
                    }

                    MultiKey key = new MultiKey(values);
                    keys[r++] = key;
                }

                final DefaultIndex<MultiKey> index = new DefaultIndex<MultiKey>(keys);
                return index;
            }

            @Override
            public Index<String> buildColumnIndex() {
                return columnIndex;
            }
        });
    }

    @Override
    public DataFrame join(Series series, String[] columns) {
        MutableSeries<Integer,V>[] array = new MutableSeries[getColumnCount() + 1];
        Object[] columnLabels = new Object[getColumnCount() + 1];
        int c = 0;
        for (String column : columns()) {
            columnLabels[c] = column;
            array[c++] = (MutableSeries<Integer,V>)this.series[columnIndex.getAddress(column)];
        }
        columnLabels[c] = series.getName();
        array[c++] = new JoinSeries(series, columns);

        return new DefaultDataFrame<V>(rowIndex, new DefaultIndex(columnLabels), array);
    }

    public DefaultDataFrame<V> removeColumns(String... columns) {
        return new DefaultDataFrame<V>(rowIndex, columnIndex.remove(columns), series);
    }

    private class JoinSeries extends AbstractSeries<Integer,V> implements MutableSeries<Integer,V> {
        private Series<MultiKey,V> series;
        private String[] columns;

        private JoinSeries(Series<MultiKey,V> series, String[] columns) {
            assert series != null;

            this.series = series;
            this.columns = columns;
        }

        @Override
        public Object getName() {
            return series.getName();
        }

        @Override
        public Class getType() {
            return series.getType();
        }

        @Override
        public V get(Integer key) {
            final Object[] values = new Object[columns.length];
            int c = 0;
            for (String column : columns) {
                values[c++] = getValueAt(key, column);
            }

            MultiKey multiKey = new MultiKey(values);

            return series.get(multiKey);
        }

        @Override
        public Integer getKey(int i) {
            return null;
        }

        @Override
        public void set(Integer key, V value) {

        }

        @Override
        public Iterable<Integer> keys() {
            return DefaultDataFrame.this.rows();
        }

        @Override
        public <L> Series<L, V> reindex(L... keys) {
            return null;
        }

        @Override
        public Series<Integer, V> head(int count) {
            return null;
        }

        @Override
        public Series<Integer, V> tail(int count) {
            return null;
        }
    }

    private class ColumnSeries extends AbstractSeries<Integer,V> {
        private String column;
        private Index index;

        private ColumnSeries(String column, Index index) {
            assert column != null;

            this.column = column;
            this.index = index;
        }

        @Override
        public Object getName() {
            return column;
        }

        @Override
        public Class getType() {
            return getColumnClass(column);
        }

        @Override
        public V get(Integer key) {
            return getValueAt(rowIndex.getKey(index.getAddress(key)), column);
        }

        @Override
        public Integer getKey(int i) {
            return rowIndex.getKey(i);
        }

        @Override
        public Iterable<Integer> keys() {
            return index.keys();
        }

        @Override
        public <L> Series<L, V> reindex(L... keys) {
            return null;
        }

        @Override
        public Series<Integer, V> head(int count) {
            return null;
        }

        @Override
        public Series<Integer, V> tail(int count) {
            return null;
        }
    }

    private class ColumnSeries2 extends AbstractSeries<Integer,V> {
        private final String column;
        private final Series<Integer, V> s;

        private ColumnSeries2(String column) {
            assert column != null;

            this.column = column;

            final Integer address = columnIndex.getAddress(column);
            if(address != null) {
                s = series[address];
            } else {
                s = null;
            }
        }

        @Override
        public Object getName() {
            return column;
        }

        @Override
        public Class getType() {
            return s != null ? s.getType() : null;
        }

        @Override
        public V get(Integer key) {
            return s != null ? s.get(key) : null;
        }

        @Override
        public Integer getKey(int i) {
            return s != null ? s.getKey(i) : null;
        }

        @Override
        public Iterable<Integer> keys() {
            return rows();
        }

        @Override
        public <L> Series<L, V> reindex(L... keys) {
            return null;
        }

        @Override
        public Series<Integer, V> head(int count) {
            return null;
        }

        @Override
        public Series<Integer, V> tail(int count) {
            return null;
        }
    }
}
