/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.dataframe;

import com.macrofocus.molap.index.DefaultIndex;
import com.macrofocus.molap.index.Index;
import com.macrofocus.molap.index.MultiKey;
import com.macrofocus.molap.series.AbstractSeries;
import com.macrofocus.molap.series.Series;

import java.util.HashSet;
import java.util.Set;

/**
 * An indexed data frame.
 *
 * @param <R> the type of row keys
 * @param <V> the type of column keys
 */
public class IndexedDataFrame<R,V> extends AbstractDataFrame<R,Object,V> {
    private final Index<Object> columnIndex;
    private final Index<R> rowIndex;
    private final Series<R,V>[] series;

    public IndexedDataFrame(Series<R,V>... series) {
        this.series = series;

        Set<R> rows = new HashSet<R>();
        for (Series<R, V> s : series) {
            for (R r : s.keys()) {
                rows.add(r);
            }
        }

        Object[] columns = new Object[series.length];
        for (int i = 0; i < series.length; i++) {
            Series<R, V> s = series[i];
            if(s.getName() != null) {
                columns[i] = s.getName();
            } else {
                columns[i] = i;
            }
        }

        R[] array = (R[])rows.toArray();

        rowIndex = new DefaultIndex<R>(array);
        columnIndex = new DefaultIndex<Object>(columns);
    }

    public IndexedDataFrame(Index<R> rowIndex, Index columnIndex, Series<R, V>[] series) {
        this.rowIndex = rowIndex;
        this.columnIndex = columnIndex;
        this.series = series;
    }

    public Index<Object> getColumnIndex() {
        return columnIndex;
    }

    public Index<R> getRowIndex() {
        return rowIndex;
    }

    @Override
    public Class getRowClass(R row) {
        return null;
    }

    @Override
    public Class getColumnClass(Object column) {
        return series[columnIndex.getAddress(column)].getType();
    }

    @Override
    public Series<Object, ?> getRow(Object row) {
        return null;
    }

    @Override
    public Series<R, V> getColumn(Object column) {
        return new ColumnSeries<R>(column, rowIndex);
    }

    @Override
    public V getValueAt(R row, Object column) {
        final Integer address = columnIndex.getAddress(column);
        if(address != null) {
            return series[address].get(row);
        } else {
            return null;
        }
    }

    @Override
    public DataFrame<Object, R, ?> append(Series<Object, ?> series) {
        return null;
    }

    @Override
    public Iterable<R> rows() {
        return rowIndex.keys();
    }

    @Override
    public Iterable<Object> columns() {
        return columnIndex.keys();
    }

    @Override
    public R getRowKey(int index) {
        return rowIndex.getKey(index);
    }

    @Override
    public Object getColumnKey(int index) {
        return columnIndex.getKey(index);
    }

    @Override
    public int getRowAddress(R row) {
        return rowIndex.getAddress(row);
    }

    @Override
    public int getColumnAddress(Object column) {
        return columnIndex.getAddress(column);
    }

    @Override
    public int getRowCount() {
        return rowIndex.getSize();
    }

    @Override
    public int getColumnCount() {
        return columnIndex.getSize();
    }

    @Override
    public DataFrame<MultiKey, Object, V> reindexRows(final Object... columns) {
        int r = 0;
        final MultiKey[] keys = new MultiKey[getRowCount()];
        for(R row: rows()) {
            final Object[] values = new Object[columns.length];
            int c = 0;
            for (Object column : columns) {
                values[c++] = getValueAt(row, column);
            }

            MultiKey key = new MultiKey(values);
            keys[r++] = key;
        }

        final DefaultIndex<MultiKey> index = new DefaultIndex<MultiKey>(keys);

        Series<MultiKey,V>[] series = new Series[getColumnCount()];
        int c = 0;
        for (Object column : columns()) {
            series[c++] = new ColumnSeries<MultiKey>(column, index);
        }

        return new IndexedDataFrame<MultiKey,V>(index, columnIndex, series);
    }

    @Override
    public DataFrame join(Series series, Object[] columns) {
        return null;
    }

    public class ColumnSeries<K> extends AbstractSeries<K,V> {
        private Object column;
        private Index index;

        private ColumnSeries(Object column, Index index) {
            assert column != null;

            this.column = column;
            this.index = index;
        }

        @Override
        public Object getName() {
            return column;
        }

        @Override
        public Class getType() {
            return getColumnClass(column);
        }

        @Override
        public V get(K key) {
            final Integer address = index.getAddress(key);
            if(address != null) {
                return getValueAt(rowIndex.getKey(address), column);
            } else {
//                System.err.println("Could not find " + key);
                return null;
            }
        }

        @Override
        public K getKey(int i) {
            return null;
        }

        @Override
        public Iterable<K> keys() {
            return index.keys();
        }

        @Override
        public <L> Series<L, V> reindex(L... keys) {
            return null;
        }

        @Override
        public Series<K, V> head(int count) {
            return null;
        }

        @Override
        public Series<K, V> tail(int count) {
            return null;
        }
    }
}
