/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.dataframe;

import com.macrofocus.molap.index.Index;
import com.macrofocus.molap.series.AbstractSeries;
import com.macrofocus.molap.series.Series;

/**
 * A reindexed data frame.
 *
 * @param <R> the rexindexed type of rows
 * @param <C> the reindexed type of columns
 * @param <V> the type of values
 * @param <X> the original type of rows
 * @param <Y> the original type of columns
 */
public class ReIndexedDataFrame<R,C,V,X,Y> extends AbstractDataFrame<R,C,V> {
    private Index<C> columnIndex;
    private Index<R> rowIndex;

    private final DataFrame<X,Y,V> dataFrame;
    private final ReIndexRecipe<R,C> recipe;

    private final DataFrameListener<X,Y> dataFrameListener = new DataFrameListener<X, Y>() {
        @Override
        public void dataFrameChanged(DataFrameEvent<X, Y> event) {
            rowIndex = null;
            columnIndex = null;

            notifyDataFrameChanged(new DataFrameEvent<R, C>(null, null, true));
        }
    };

    public ReIndexedDataFrame(DataFrame<X, Y, V> dataFrame, ReIndexRecipe<R,C> recipe) {
        this.dataFrame = dataFrame;

        this.recipe = recipe;

        dataFrame.addWeakDataFrameListener(dataFrameListener);
    }

    @Override
    public Class getRowClass(R row) {
        return dataFrame.getRowClass(dataFrame.getRowKey(getRowIndex().getAddress(row)));
    }

    @Override
    public Class getColumnClass(C column) {
        return dataFrame.getColumnClass(dataFrame.getColumnKey(getColumnIndex().getAddress(column)));
    }

    @Override
    public V getValueAt(R row, C column) {
        final Integer r = getRowIndex().getAddress(row);
        final Integer c = getColumnIndex().getAddress(column);
        if(r != null && c != null) {
            return dataFrame.getValueAt(dataFrame.getRowKey(r), dataFrame.getColumnKey(c));
        } else {
            return null;
        }
    }

    @Override
    public Series<C, ?> getRow(R row) {
        return null;
    }

    @Override
    public Series<R, V> getColumn(C column) {
//        return new ReIndexedSeries<R,V,X>(dataFrame.getColumn(dataFrame.getColumnKey(getColumnIndex().getAddress(column))), new ReIndexedSeries.ReIndexRecipe<R>() {
//            @Override
//            public Index<R> buildIndex() {
//                return getRowIndex();
//            }
//        });
        return new ColumnSeries2(column);
    }

    @Override
    public DataFrame<C, R, ?> append(Series<C, ?> series) {
        return null;
    }

    @Override
    public Iterable<R> rows() {
        return getRowIndex().keys();
    }

    @Override
    public Iterable<C> columns() {
        return getColumnIndex().keys();
    }

    @Override
    public R getRowKey(int index) {
        return getRowIndex().getKey(index);
    }

    @Override
    public C getColumnKey(int index) {
        return getColumnIndex().getKey(index);
    }

    @Override
    public int getRowAddress(R row) {
        return getRowIndex().getAddress(row);
    }

    @Override
    public int getColumnAddress(C column) {
        return getColumnIndex().getAddress(column);
    }

    @Override
    public int getRowCount() {
        return getRowIndex().getSize();
    }

    @Override
    public int getColumnCount() {
        return getColumnIndex().getSize();
    }

    @Override
    public DataFrame join(Series series, C[] columns) {
        return null;
    }

    public Index<R> getRowIndex() {
        if(rowIndex == null) {
            rowIndex = recipe.builRowdIndex();
        }
        return rowIndex;
    }

    public Index<C> getColumnIndex() {
        if(columnIndex == null) {
            columnIndex = recipe.buildColumnIndex();
        }

        assert columnIndex != null;

        return columnIndex;
    }

    public interface ReIndexRecipe<R,C> {
        Index<R> builRowdIndex();

        Index<C> buildColumnIndex();
    }

    private class ColumnSeries2 extends AbstractSeries<R,V> {
        private final C column;
        private final Series<X, V> s;

        private ColumnSeries2(C column) {
            assert column != null;

            this.column = column;

            final Integer address = getColumnIndex().getAddress(column);
            if(address != null) {
                s = dataFrame.getColumn(dataFrame.getColumnKey(address));
            } else {
                s = null;
            }
        }

        @Override
        public Object getName() {
            return column;
        }

        @Override
        public Class getType() {
            return s != null ? s.getType() : null;
        }

        @Override
        public V get(R key) {
            final Integer r = getRowIndex().getAddress(key);
            if(r != null) {
                return s.get(dataFrame.getRowKey(r));
            } else {
                return null;
            }
        }

        @Override
        public R getKey(int i) {
            return getRowKey(i);
        }

        @Override
        public Iterable<R> keys() {
            return rows();
        }

        @Override
        public <L> Series<L, V> reindex(L... keys) {
            return null;
        }

        @Override
        public Series<R, V> head(int count) {
            return null;
        }

        @Override
        public Series<R, V> tail(int count) {
            return null;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!super.equals(o)) return false;

            ColumnSeries2 that = (ColumnSeries2) o;

            if (!column.equals(that.column)) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = super.hashCode();
            result = 31 * result + column.hashCode();
            return result;
        }
    }

}
