/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.dataframe;

import com.macrofocus.molap.index.DefaultIndex;
import com.macrofocus.molap.index.Index;
import com.macrofocus.molap.series.Series;
import com.macrofocus.selection.MutableSelection;
import com.macrofocus.selection.SelectionEvent;
import com.macrofocus.selection.SelectionListener;

import java.util.ArrayList;
import java.util.List;

/**
 * A row filterable data frame. Only the selected rows will be made visible by this data frame.
 *
 * @param <R> the type of row keys
 * @param <C> the type of column keys
 * @param <V> the type of values
 */
public class SelectionDataFrame<R,C,V> extends AbstractDataFrame<R,C,V> {
    private final AbstractDataFrame<R,C,V> dataFrame;
    private final MutableSelection<R> filter;

    private Index<R> rowIndex;

    private final DataFrameListener<R, C> dataFrameListener = new DataFrameListener<R, C>() {
        @Override
        public void dataFrameChanged(DataFrameEvent<R, C> event) {
            rowIndex = null;

            notifyDataFrameChanged(new DataFrameEvent<R, C>(null, null, true));
        }
    };

    private final SelectionListener<R> selectionListener = new SelectionListener<R>() {
        @Override
        public void selectionChanged(SelectionEvent<R> event) {
            rowIndex = null;

            notifyDataFrameChanged(new DataFrameEvent<R, C>(null, null, true));
        }
    };

    public SelectionDataFrame(AbstractDataFrame<R, C, V> dataFrame, MutableSelection<R> filter) {
        this.dataFrame = dataFrame;
        this.filter = filter;

        dataFrame.addWeakDataFrameListener(dataFrameListener);
        filter.addWeakSelectionListener(selectionListener);
    }

    @Override
    public Index<R> getRowIndex() {
        if (rowIndex == null) {
            List<R> rows = new ArrayList<>();
            for (R row : dataFrame.rows()) {
                if (filter.isSelected(row)) {
                    rows.add(row);
                }
            }

            rowIndex = new DefaultIndex<R>((R[]) rows.toArray());
        }
        return rowIndex;
    }

    @Override
    public Index<C> getColumnIndex() {
        return dataFrame.getColumnIndex();
    }

    @Override
    public Class getRowClass(R r) {
        return dataFrame.getRowClass(r);
    }

    @Override
    public Class getColumnClass(C c) {
        return dataFrame.getColumnClass(c);
    }

    @Override
    public V getValueAt(R r, C c) {
        return dataFrame.getValueAt(r, c);
    }

    @Override
    public Series<C, ?> getRow(R r) {
        return dataFrame.getRow(r);
    }

    @Override
    public Iterable<R> rows() {
        return getRowIndex().keys();
    }

    @Override
    public Iterable<C> columns() {
        return dataFrame.columns();
    }

    @Override
    public R getRowKey(int index) {
        return getRowIndex().getKey(index);
    }

    @Override
    public C getColumnKey(int index) {
        return dataFrame.getColumnKey(index);
    }

    @Override
    public int getRowAddress(R r) {
        return getRowIndex().getAddress(r);
    }

    @Override
    public int getColumnAddress(C c) {
        return dataFrame.getColumnAddress(c);
    }

    @Override
    public int getRowCount() {
        return getRowIndex().getSize();
    }

    @Override
    public int getColumnCount() {
        return dataFrame.getColumnCount();
    }

    @Override
    public DataFrame<C, R, ?> append(Series<C, ?> series) {
        return null;
    }

    @Override
    public DataFrame join(Series series, C[] cs) {
        return null;
    }
}
