/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.dataframe.swing;

import com.macrofocus.molap.dataframe.AbstractDataFrame;
import com.macrofocus.molap.dataframe.DataFrame;
import com.macrofocus.molap.dataframe.DataFrameEvent;
import com.macrofocus.molap.dataframe.MutableDataFrame;
import com.macrofocus.molap.index.DefaultIndex;
import com.macrofocus.molap.index.Index;
import com.macrofocus.molap.index.IntegerIndex;
import com.macrofocus.molap.index.MultiKey;
import com.macrofocus.molap.series.AbstractSeries;
import com.macrofocus.molap.series.Series;

import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableModel;

/**
 * Converts a TableModel into a DataFrame.
 */
public class TableModelDataFrame extends AbstractDataFrame<Integer,String,Object> implements MutableDataFrame<Integer,String,Object> {
    private final TableModel tableModel;

    private Index<Integer> rowIndex = null;
    private Index<String> columnIndex = null;

    public TableModelDataFrame(TableModel tableModel) {
        this.tableModel = tableModel;

        tableModel.addTableModelListener(new TableModelListener() {
            @Override
            public void tableChanged(TableModelEvent e) {
                rowIndex = null;
                columnIndex = null;

                notifyDataFrameChanged(new DataFrameEvent<Integer, String>(null, null, true));
            }
        });
    }

    public TableModel getTableModel() {
        return tableModel;
    }

    @Override
    public Class getRowClass(Integer row) {
        return Object.class;
    }

    @Override
    public Class getColumnClass(String column) {
        return tableModel.getColumnClass(getColumnIndex().getAddress(column));
    }

    @Override
    public Object getValueAt(Integer row, String column) {
        return tableModel.getValueAt(row, getColumnIndex().getAddress(column));
    }

    @Override
    public void setValueAt(Integer row, String column, Object value) {
        tableModel.setValueAt(value, row, getColumnIndex().getAddress(column));
    }

    @Override
    public Series<String, ?> getRow(Integer row) {
        return null;
    }

    @Override
    public Series<Integer, Object> getColumn(String column) {
        return new ColumnSeries(column);
    }

    @Override
    public DataFrame<String, Integer, ?> append(Series<String, ?> series) {
        return null;
    }

    @Override
    public Iterable<Integer> rows() {
        return getRowIndex().keys();
    }

    @Override
    public Iterable<String> columns() {
        return getColumnIndex().keys();
    }

    @Override
    public Integer getRowKey(int index) {
        return getRowIndex().getKey(index);
    }

    @Override
    public String getColumnKey(int index) {
        return getColumnIndex().getKey(index);
    }

    @Override
    public int getRowAddress(Integer row) {
        return getRowIndex().getAddress(row);
    }

    @Override
    public int getColumnAddress(String column) {
        return getColumnIndex().getAddress(column);
    }

    @Override
    public int getRowCount() {
        return tableModel.getRowCount();
    }

    @Override
    public int getColumnCount() {
        return tableModel.getColumnCount();
    }

    @Override
    public DataFrame<MultiKey, String, Object> reindexRows(String... columns) {
        return null;
    }

    @Override
    public DataFrame join(Series series, String[] columns) {
        return null;
    }

    @Override
    public Index<Integer> getRowIndex() {
        if(rowIndex == null) {
            rowIndex = new IntegerIndex(0, tableModel.getRowCount() - 1);
        }
        return rowIndex;
    }

    @Override
    public Index<String> getColumnIndex() {
        if(columnIndex == null) {
            String[] names = new String[tableModel.getColumnCount()];

            for(int c = 0; c < tableModel.getColumnCount(); c++) {
                names[c] = tableModel.getColumnName(c);
            }

            columnIndex = new DefaultIndex<String>(names);
        }

        return columnIndex;
    }

    private class ColumnSeries extends AbstractSeries<Integer,Object> {
        private final String column;

        private ColumnSeries(String column) {
            assert column != null;

            this.column = column;
        }

        @Override
        public Object getName() {
            return column;
        }

        @Override
        public Class getType() {
            return getColumnClass(column);
        }

        @Override
        public Object get(Integer key) {
            return getValueAt(key, column);
        }

        @Override
        public Integer getKey(int i) {
            return getRowKey(i);
        }

        @Override
        public Iterable<Integer> keys() {
            return rows();
        }

        @Override
        public <L> Series<L, Object> reindex(L... keys) {
            return null;
        }

        @Override
        public Series<Integer, Object> head(int count) {
            return null;
        }

        @Override
        public Series<Integer, Object> tail(int count) {
            return null;
        }
    }
}
