/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */
package com.macrofocus.molap.table;

public abstract class AbstractColumn implements Column {
    @Override
    public Column add(double value) {
        return new ObjectColumn(this, value, add);
    }

    @Override
    public Column add(final Column column) {
        return new ColumnColumn(this, column, add);
    }

    @Override
    public Column subtract(double value) {
        return new ObjectColumn(this, value, subtract);
    }

    @Override
    public Column substract(final Column column) {
        return new ColumnColumn(this, column, subtract);
    }

    @Override
    public Column multiply(double value) {
        return new ObjectColumn(this, value, multiply);
    }

    @Override
    public Column multiply(final Column column) {
        return new ColumnColumn(this, column, multiply);
    }

    @Override
    public Column divide(double value) {
        return new ObjectColumn(this, value, divide);
    }

    @Override
    public Column divide(final Column column) {
        return new ColumnColumn(this, column, divide);
    }

    @Override
    public Column pow(double value) {
        return new ObjectColumn(this, value, pow);
    }

    @Override
    public Column pow(final Column column) {
        return new ColumnColumn(this, column, pow);
    }

    private static class ObjectColumn extends AbstractColumn {
        private final Column c1;
        private final Object c2;
        private final Operation operation;

        private ObjectColumn(Column c1, Object c2, Operation operation) {
            this.c1 = c1;
            this.c2 = c2;
            this.operation = operation;
        }

        @Override
        public String getName() {
            return operation.name(c1.getName(), c2);
        }

        @Override
        public Class getType() {
            return operation.getType();
        }

        @Override
        public int getRowCount() {
            return c1.getRowCount();
        }

        @Override
        public Row getRow(int row) {
            return c1.getRow(row);
        }

        @Override
        public Object getValueAt(int row) {
            return operation.compute(c1.getValueAt(row), c2);
        }
    }

    private static class ColumnColumn extends AbstractColumn {
        private final Column c1;
        private final Column c2;
        private final Operation operation;

        private ColumnColumn(Column c1, Column c2, Operation operation) {
            this.c1 = c1;
            this.c2 = c2;
            this.operation = operation;
        }

        @Override
        public String getName() {
            return operation.name(c1.getName(), c2.getName());
        }

        @Override
        public Class getType() {
            return operation.getType();
        }

        @Override
        public int getRowCount() {
            return c1.getRowCount();
        }

        @Override
        public Row getRow(int row) {
            return c1.getRow(row);
        }

        @Override
        public Object getValueAt(int row) {
            return operation.compute(c1.getValueAt(row), c2.getValueAt(row));
        }
    }

    private static interface Operation {
        String name(String n1, Object n2);

        Class getType();

        Object compute(Object a, Object b);
    }

    private static abstract class AbstractOperation implements Operation {
        private final Class type;

        protected AbstractOperation(Class type) {
            this.type = type;
        }

        @Override
        public Class getType() {
            return type;
        }
    }

    private static final Operation add = new AbstractOperation(Double.class) {
        @Override
        public String name(String n1, Object n2) {
            return n1 + "+" + n2;
        }

        @Override
        public Object compute(Object a, Object b) {
            if(a != null && b != null && a instanceof Number && b instanceof Number) {
                return ((Number)a).doubleValue() + ((Number)b).doubleValue();
            } else {
                return null;
            }
        }
    };

    private static final Operation subtract = new AbstractOperation(Double.class) {
        @Override
        public String name(String n1, Object n2) {
            return n1 + "-" + n2;
        }

        @Override
        public Object compute(Object a, Object b) {
            if(a != null && b != null && a instanceof Number && b instanceof Number) {
                return ((Number)a).doubleValue() - ((Number)b).doubleValue();
            } else {
                return null;
            }
        }
    };

    private static final Operation multiply = new AbstractOperation(Double.class) {
        @Override
        public String name(String n1, Object n2) {
            return n1 + "*" + n2;
        }

        @Override
        public Object compute(Object a, Object b) {
            if(a != null && b != null && a instanceof Number && b instanceof Number) {
                return ((Number)a).doubleValue() * ((Number)b).doubleValue();
            } else {
                return null;
            }
        }
    };

    private static final Operation divide = new AbstractOperation(Double.class) {
        @Override
        public String name(String n1, Object n2) {
            return n1 + "/" + n2;
        }

        @Override
        public Object compute(Object a, Object b) {
            if(a != null && b != null && a instanceof Number && b instanceof Number) {
                return ((Number)a).doubleValue() / ((Number)b).doubleValue();
            } else {
                return null;
            }
        }
    };

    private static final Operation pow = new AbstractOperation(Double.class) {
        @Override
        public String name(String n1, Object n2) {
            return "pow(" + n1 + "," + n2 + ")";
        }

        @Override
        public Object compute(Object a, Object b) {
            if(a != null && b != null && a instanceof Number && b instanceof Number) {
                return Math.pow(((Number)a).doubleValue(), ((Number)b).doubleValue());
            } else {
                return null;
            }
        }
    };
}
