/*
 * Copyright (c) 2014 Macrofocus GmbH. All Rights Reserved.
 */

import com.macrofocus.molap.aggregates.AggregateDataFrame;
import com.macrofocus.molap.aggregates.aggregation.Aggregation;
import com.macrofocus.molap.aggregates.aggregation.StdDevAggregation;
import com.macrofocus.molap.dataframe.DataFrame;
import com.macrofocus.molap.dataframe.DataFrameFactory;
import com.macrofocus.molap.series.Series;

import java.io.PrintStream;
import java.util.LinkedList;
import java.util.List;

public class Fruits {
    public static void main(String[] args) {
        final List<QueryEntry> queries = new LinkedList<>();

        DataFrame<Integer, String, Object> dataFrame = DataFrameFactory.fromRowMajorArray(
                new String[]{"Item", "Origin", "Quantity", "Price"},
                new Class[]{String.class, String.class, Integer.class, Double.class},
                new Object[][]{
                        {"Apple", "Switzerland", 12, 112.0},
                        {"Pear", "Switzerland", 24, 125.0},
                        {"Pear", "Switzerland", 6, 130.0},
                        {"Tomato", "Switzerland", 50, 122.0},
                        {"Tomato", "Italy", 40, 44.0},
                        {"Tomato", "Italy", 100, 80.0}
                }
        );

        // Overall quantity and revenue
        Aggregation quantity = dataFrame.getSum("Quantity");
        Aggregation revenue = dataFrame.getSum("Price").as("Revenue");

        AggregateDataFrame<String> apexQuery = dataFrame.aggregate(quantity, revenue);
        queries.add(new QueryEntry(apexQuery,
                "Querying the highest level of abstraction",
                "Aggregation quantity = dataFrame.getSum(\"Quantity\");\n" +
                "Aggregation revenue = dataFrame.getSum(\"Price\").as(\"Revenue\");\n" +
                "AggregateDataFrame<String> apexQuery = dataFrame.aggregate(quantity, revenue);",
                "Overall quantity and revenue",
                "A cube at the highest level of abstraction is the apex cuboid. When a query is placed, this the the default cuboid being used."));

        // Quantities and revenues by item
        AggregateDataFrame<String> byItemQuery = dataFrame.aggregate(quantity, revenue)
                .drillDown("Item");
        queries.add(new QueryEntry(byItemQuery,
                "Drilling down one level",
                "AggregateDataFrame<String> byItemQuery = dataFrame.aggregate(quantity, revenue)" +
                "        .drillDown(\"Item\");",
                "Quantities and revenues by item",
                "Drill Down/Up allows the user to navigate among levels of data ranging from the most summarized (up) to the most detailed (down)."));

        // Quantities and revenues by item ordered by quantity
        AggregateDataFrame<String> orderQuery = dataFrame.aggregate(quantity, revenue)
                .drillDown("Item").order(revenue);
        queries.add(new QueryEntry(orderQuery,
                "Order by a specific value",
                "AggregateDataFrame<String> orderQuery = dataFrame.aggregate(quantity, revenue)" +
                "        .drillDown(\"Item\").order(revenue);",
                "Quantities and revenues by item ordered by quantity",
                "Sorting can be defined through one or more aggregation value."));

        // Quantities and revenues by item
        AggregateDataFrame<String> byOriginQuery = apexQuery.drillDown("Origin");
        queries.add(new QueryEntry(byOriginQuery,
                "Deriving an existing query",
                "AggregateDataFrame<String> byOriginQuery = apexQuery.drillDown(\"Origin\");",
                "Quantities and revenues by origin",
                "An existing query can be further refined."));

        // Quantities and revenues by item and origin
        AggregateDataFrame<String> byItemOriginQuery = byItemQuery.drillDown("Origin");
        queries.add(new QueryEntry(byItemOriginQuery,
                "Drilling down multiple levels",
                "AggregateDataFrame<String> byItemOriginQuery = byItemQuery.drillDown(\"Origin\");",
                "Quantities and revenues by item and origin",
                "Drilling can be done on any number of levels, the cube will adapt automatically."));

        // Quantities and revenues by origin and item
        AggregateDataFrame<String> byOriginItemQuery = byOriginQuery.drillDown("Item");
        queries.add(new QueryEntry(byOriginItemQuery,
                "Alternate group by",
                "AggregateDataFrame<String> byOriginItemQuery = byOriginQuery.drillDown(\"Item\");",
                "Quantities and revenues by origin and item",
                "A different drilling path should lead to the same results!"));

        // Revenues by item and origin (pivot)
        AggregateDataFrame<String> pivotQuery = byItemOriginQuery.pivot(revenue);
        queries.add(new QueryEntry(pivotQuery,
                "Pivoting",
                "AggregateDataFrame<String> pivotQuery = byItemOriginQuery.pivot(revenue);",
                "Revenues by item and origin (pivot)",
                "Pivot allows an analyst to rotate the cube in space to see its various faces."));

        // Quantities and revenues by item with totals
        AggregateDataFrame<String> byItemWithTotalQuery = dataFrame.aggregate(quantity, revenue)
                .on(byItemQuery.getCuboid(), byItemQuery.getCuboid().drillUp());
        queries.add(new QueryEntry(byItemWithTotalQuery,
                "Adding summary values",
                "AggregateDataFrame<String> byItemWithTotalQuery = dataFrame.aggregate(quantity, revenue)\n" +
                "        .on(byItemQuery.getCuboid(), byItemQuery.getCuboid().drillUp());",
                "Quantities and revenues by item with totals",
                "Summary information (e.g. totals) can be added by navigating up the the cuboid hierarchy."));

        // Quantities and revenues by all combinations
        AggregateDataFrame<String> cubeQuery = dataFrame.aggregate(quantity, revenue)
                .on(byItemOriginQuery.getCuboid(),
                        byItemQuery.getCuboid(),
                        byOriginQuery.getCuboid(),
                        apexQuery.getCuboid());
        queries.add(new QueryEntry(cubeQuery,
                "All cuboids",
                "AggregateDataFrame<String> cubeQuery = dataFrame.aggregate(quantity, revenue)\n" +
                "        .on(byItemOriginQuery.getCuboid(),\n" +
                "                byItemQuery.getCuboid(),\n" +
                "                byOriginQuery.getCuboid(),\n" +
                "                apexQuery.getCuboid());",
                "Quantities and revenues by all combinations",
                "A view of all the possible combinations provides interesting analytic possibilities."));

        // Mean and standard deviation of unit price by item
        Series unitPriceColumn = dataFrame.getColumn("Price")
                .divide(dataFrame.getColumn("Quantity"));
        Series squareUnitPriceColumn = unitPriceColumn.pow(2.0)
                .multiply(dataFrame.getColumn("Quantity"));
        Aggregation meanPrice = dataFrame.getMean("Price");
        Aggregation stdDevPrice = dataFrame.getStdDev("Price");
        Aggregation meanUnitPrice = revenue.dividedBy(quantity)
                .as("Mean unit price");
        Aggregation stdDevUnitPrice = new StdDevAggregation(squareUnitPriceColumn,
                dataFrame.getColumn("Price"), dataFrame.getColumn("Quantity"))
                .as("Standard deviation of unit price");
        AggregateDataFrame<String> statQuery = dataFrame
                .aggregate(meanPrice, stdDevPrice, meanUnitPrice, stdDevUnitPrice)
                .drillDown("Item")
                .drillDown("Origin");
        queries.add(new QueryEntry(statQuery,
                "Advanced statistics",
                "Series unitPriceColumn = dataFrame.getColumn(\"Price\")\n" +
                "        .divide(dataFrame.getColumn(\"Quantity\"));\n" +
                "Series squareUnitPriceColumn = unitPriceColumn.pow(2.0)\n" +
                "        .multiply(dataFrame.getColumn(\"Quantity\"));\n" +
                "Aggregation meanPrice = dataFrame.getMean(\"Price\");\n" +
                "Aggregation stdDevPrice = dataFrame.getStdDev(\"Price\");\n" +
                "Aggregation meanUnitPrice = revenue.dividedBy(quantity)\n" +
                "        .as(\"Mean unit price\");\n" +
                "Aggregation stdDevUnitPrice = new StdDevAggregation(squareUnitPriceColumn,\n" +
                "        dataFrame.getColumn(\"Price\"), dataFrame.getColumn(\"Quantity\"))\n" +
                "        .as(\"Standard deviation of unit price\");\n" +
                "AggregateDataFrame<String> statQuery = dataFrame\n" +
                "        .aggregate(meanPrice, stdDevPrice, meanUnitPrice, stdDevUnitPrice)\n" +
                "        .drillDown(\"Item\")\n" +
                "        .drillDown(\"Origin\");",
                "Mean and standard deviation of unit price by item",
                "The cube allow for computing advanced statistics without having to compute everything at every level."));

        PrintStream terminal = null;
        PrintStream html = null;
        PrintStream docBook = System.out;
        for (QueryEntry queryEntry : queries) {
            if(docBook != null) {
                docBook.println("<sect1>");
            }

            try {
                if(queryEntry.getTitle() != null) {
                    if(terminal != null) {
                        terminal.println(queryEntry.getTitle());
                    }
                    if(html != null) {
                        html.println("<h2>" + queryEntry.getTitle() + "</h2>");
                    }
                    if(docBook != null) {
                        docBook.println("<title>" + queryEntry.getTitle() + "</title>");
                    }
                }

                if(queryEntry.getDescription() != null) {
                    if(terminal != null) {
                        terminal.println(queryEntry.getDescription());
                    }
                    if(html != null) {
                        html.println("<p>" + queryEntry.getDescription() + "</p>");
                    }
                    if(docBook != null) {
                        docBook.println("<para>" + queryEntry.getDescription() + "</para>");
                    }
                }

                if(queryEntry.getCode() != null) {
                    if(terminal != null) {
                        terminal.println(queryEntry.getCode());
                    }
                    if(html != null) {
                        html.println("<pre class=\"programlisting\">" + queryEntry.getCode() + "</pre>");
                    }
                    if(docBook != null) {
                        docBook.println("<programlisting language=\"java\">" + queryEntry.getCode() + "</programlisting>");
                    }
                }

                DataFrame query = queryEntry.getQuery();
                if(terminal != null) {
                    query.print(terminal, queryEntry.getCaption(), false);
                }
                if(html != null) {
                    query.print(html, queryEntry.getCaption(), true);
                }
                if(docBook != null) {
                    query.print(docBook, queryEntry.getCaption(), true);
                }
            } catch (Exception e) {
                throw new RuntimeException("Error executing " + queryEntry.getTitle(), e);
            } finally {
                if(docBook != null) {
                    docBook.println("</sect1>");
                }
            }
        }
    }

    private static class QueryEntry {
        private final DataFrame query;
        private final String title;
        private final String code;
        private final String caption;
        private final String description;

        private QueryEntry(DataFrame query, String title, String code, String caption, String description) {
            this.query = query;
            this.title = title;
            this.code = code.replaceAll("<", "&lt;").replaceAll(">", "&gt;");
            this.caption = caption;
            this.description = description;
        }

        public DataFrame getQuery() {
            return query;
        }

        public String getTitle() {
            return title;
        }

        public String getCode() {
            return code;
        }

        public String getCaption() {
            return caption;
        }

        public String getDescription() {
            return description;
        }
    }
}
